/**
 *	Copyright (C) 2011-2016 Docuteam GmbH
 *
 *	This program is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 3
 *	as published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package ch.docuteam.aipcreatorETH.ingest;

import gov.loc.mets.*;
import gov.loc.mets.MetsType.FileSec.FileGrp;

import static ch.docuteam.darc.DarcConstants.UNDERSCORE_ZIP;
import static ch.docuteam.darc.DarcConstants.ZIP_EXT;

import java.io.*;
import java.lang.reflect.InvocationTargetException;
import java.sql.*;
import java.util.*;
import java.util.Date;

import javax.swing.tree.TreePath;

import org.apache.xmlbeans.XmlObject;
import org.dom4j.DocumentHelper;
import org.dom4j.Element;
import org.dom4j.Node;
import org.dom4j.io.OutputFormat;
import org.dom4j.io.XMLWriter;
import org.jdesktop.swingx.treetable.MutableTreeTableNode;

import ch.docuteam.darc.exceptions.*;
import ch.docuteam.darc.ingest.AIPCreator;
import ch.docuteam.darc.mdconfig.LevelOfDescription;
import ch.docuteam.darc.mets.Document;
import ch.docuteam.darc.mets.Document.Mode;
import ch.docuteam.darc.mets.structmap.*;
import ch.docuteam.darc.mets.structmap.NodeAbstract.SubmitStatus;
import ch.docuteam.tools.file.FileFilter;
import ch.docuteam.tools.file.FileUtil;
import ch.docuteam.tools.file.Zipper;
import ch.docuteam.tools.file.exception.FileUtilExceptionListException;
import ch.docuteam.tools.file.ssh.SFTPChannel;
import ch.docuteam.tools.id.UniqueID;
import ch.docuteam.tools.os.OperatingSystem;
import ch.docuteam.tools.out.Logger;
import de.schlichtherle.truezip.file.TFile;

import com.exlibris.core.sdk.consts.Enum;
import com.exlibris.core.sdk.formatting.DublinCore;
import com.exlibris.core.sdk.formatting.DublinCoreFactory;
import com.exlibris.digitool.common.dnx.*;
import com.exlibris.digitool.common.dnx.DnxDocumentHelper.FileFixity;
import com.exlibris.dps.sdk.deposit.IEParser;
import com.exlibris.dps.sdk.deposit.IEParserFactory;
import com.jcraft.jsch.JSchException;
import com.jcraft.jsch.SftpException;
import com.jcraft.jsch.ChannelSftp;
import com.jcraft.jsch.ChannelSftp.LsEntry;

/**
 * Implementation for the ingest workflow at ETH Zürich based on file based submission (optionally using SSH connections) and storage into an ExLibris Rosetta
 * repository system.
 * 
 * The transfer into Rosetta is managed by a separate ingest application developed by ETH Zürich. Coodination between feeder and this
 * application is done using a common database.
 * 
 * @author denis
 *
 */
public class AIPCreatorETH implements AIPCreator
{
	//	===========================================================================================
	//	========	Structure				=======================================================
	//	===========================================================================================

	//	========	Static Final Public		=======================================================

	//	========	Static Final Private	=======================================================

	static private final String					FeedbackRelativeFolderName = File.separator + "feedback";

	static private final String					OSSuffix = OperatingSystem.isWindows() ? "Win" : (OperatingSystem.isMacOSX() ? "OSX" : (OperatingSystem.isLinux() ? "Linux" : ""));


	//	TODO	maybe have this mapping in an separate file to enable dynamic changes?
	static private final Map<String, String>	MetadataNamesMapToDC = new HashMap<String, String>();
	static
	{
		MetadataNamesMapToDC.put("PID", "dc:identifier");
		MetadataNamesMapToDC.put("origination", "dc:creator");
		MetadataNamesMapToDC.put("author", "dc:creator");
		MetadataNamesMapToDC.put("creationPeriod", "dcterms:temporal");
		MetadataNamesMapToDC.put("creationPeriodNotes", "dc:description");
		MetadataNamesMapToDC.put("scopeContent", "dc:subject");
		MetadataNamesMapToDC.put("language", "dc:language");
		MetadataNamesMapToDC.put("languageNotes", "dc:description");
		MetadataNamesMapToDC.put("accessRestriction", "dc:rights");
		//	Note: accessRestrictionPeriod is handled separately! See method "generateIE(...)" further below.
		MetadataNamesMapToDC.put("accessRestrictionPeriodBaseYear", "dc:date");
		MetadataNamesMapToDC.put("accessRestrictionExplanation", "dc:description");
		//	Note: retentionPolicy is handled separately! See method "generateIE(...)" further below.
		//	Note: retentionPeriod is handled separately! See method "generateIE(...)" further below.
		MetadataNamesMapToDC.put("retentionPeriodBaseYear", "dc:available");
		MetadataNamesMapToDC.put("archivalHistory", "dc:description");
		MetadataNamesMapToDC.put("objectType", "dc:type");
		MetadataNamesMapToDC.put("extent", "dc:extent");
		MetadataNamesMapToDC.put("refCodeOld", "dc:identifier");
		MetadataNamesMapToDC.put("refCode", "dc:identifier");
		//	Note: refCodeAdmin is handled separately! See method "generateAIP(...)" further below.
		MetadataNamesMapToDC.put("reproductions", "dcterms:hasVersion");
		MetadataNamesMapToDC.put("comment", "dc:description");
		MetadataNamesMapToDC.put("institution", "dc:publisher");
		MetadataNamesMapToDC.put("arrangement", "dc:format");
		MetadataNamesMapToDC.put("accessNr", "dc:source");

		MetadataNamesMapToDC.put("appraisalAndDestruction", "dc:description");
		MetadataNamesMapToDC.put("bibliography", "dcterms:isReferencedBy");
		MetadataNamesMapToDC.put("unitTitleAdditional", "dcterms:alternative");
		MetadataNamesMapToDC.put("material", "dc:medium");
		MetadataNamesMapToDC.put("conditionsOfReproductions", "dc:description");
		MetadataNamesMapToDC.put("relatedMaterial", "dcterms:references");
		MetadataNamesMapToDC.put("descriptionLevel", "dc:description");
		MetadataNamesMapToDC.put("descriptionLevelNotes", "dc:description");
		MetadataNamesMapToDC.put("descriptionRules", "dc:description");
		MetadataNamesMapToDC.put("revisions", "dcterms:hasVersion");
		MetadataNamesMapToDC.put("responsible", "dc:rightsHolder");
		MetadataNamesMapToDC.put("involved", "dc:contributor");
		MetadataNamesMapToDC.put("staff", "dc:contributor");
		MetadataNamesMapToDC.put("projectName", "dcterms:alternative");
		MetadataNamesMapToDC.put("projectTitle", "dcterms:alternative");
		MetadataNamesMapToDC.put("project", "dc:type");
		MetadataNamesMapToDC.put("projectAbbreviation", "dc:type");
		MetadataNamesMapToDC.put("location", "dcterms:spatial");
		MetadataNamesMapToDC.put("journal", "dcterms:bibliographicCitation");
		MetadataNamesMapToDC.put("university", "dc:contributor");
		MetadataNamesMapToDC.put("institute", "dc:contributor");
		MetadataNamesMapToDC.put("compartment", "dc:subject");
		MetadataNamesMapToDC.put("method", "dc:coverage");
		MetadataNamesMapToDC.put("keyword", "dc:subject");
		MetadataNamesMapToDC.put("fundingSource", "dcterms:mediator");
		MetadataNamesMapToDC.put("status", "dcterms:conformsTo");
		MetadataNamesMapToDC.put("publisher", "dc:publisher");
		MetadataNamesMapToDC.put("doiJournal", "dcterms:isReferencedBy");
		MetadataNamesMapToDC.put("usage", "dcterms:requires");
		MetadataNamesMapToDC.put("year", "dc:date");
		MetadataNamesMapToDC.put("abstract", "dcterms:abstract");
	}

	//	========	Static Public			=======================================================

	static public boolean				IsTestRun = false;		//	ToDo: This is temporary during development
	static public boolean				IsWithoutDB = false;	//	ToDo: This is temporary during development

	//	========	Static Private			=======================================================

	//	========	Instance Public			=======================================================

	//	========	Instance Private		=======================================================

	private String						aipFolderPath;
	private String						executionId;
	private String						deeplinkPatternCollection;
	private String						deeplinkPatternIE;

	private String						ingestSubmitDir;
	private String						sshHost;
	private String						sshUser;
	private int							sshPort;
	private String						sshPathToPK;
	private String						sshPathToKnownHosts;

	private String						dbHost;
	private String						dbPort;
	private String						dbService;
	private String						dbName;
	private String						dbUserName;
	private String						dbUserPassword;

	//	===========================================================================================
	//	========	Main					=======================================================
	//	===========================================================================================

	//	===========================================================================================
	//	========	Methods					=======================================================
	//	===========================================================================================

	//	========	Static Initializer		=======================================================

	//	========	Constructors Public		=======================================================

	//	========	Constructors Private	=======================================================

	//	========	Static Public			=======================================================

	//	--------		Initializing		-------------------------------------------------------
	//	--------		Accessing			-------------------------------------------------------
	//	--------		Inquiring			-------------------------------------------------------
	//	--------		Business Ops		-------------------------------------------------------
	//	--------		Persistence			-------------------------------------------------------
	//	--------		Support				-------------------------------------------------------
	//	--------		Utilities			-------------------------------------------------------
	//	---------		Misc				-------------------------------------------------------
	//	--------		Debugging			-------------------------------------------------------
	//	---------		Temporary			-------------------------------------------------------

	//	========	Static Private			=======================================================

	//	--------		Initializing		-------------------------------------------------------
	//	--------		Accessing			-------------------------------------------------------
	//	--------		Inquiring			-------------------------------------------------------
	//	--------		Business Ops		-------------------------------------------------------
	//	--------		Persistence			-------------------------------------------------------
	//	--------		Support				-------------------------------------------------------
	//	--------		Utilities			-------------------------------------------------------
	//	---------		Misc				-------------------------------------------------------
	//	--------		Debugging			-------------------------------------------------------
	//	---------		Temporary			-------------------------------------------------------

	//	========	Instance Public			=======================================================

	//	--------		Initializing		-------------------------------------------------------
	//	--------		Accessing			-------------------------------------------------------
	//	--------		Inquiring			-------------------------------------------------------
	//	--------		Interface			-------------------------------------------------------

	/**
	 * Mandatory properties for generateAIP()/confirmStorage():
	 * - AIPCreator.aipFolderPath				Target folder for archival objects to be transferred to the Rosetta repository
	 * - AIPCreator.executionId					The ID of the feeder workflow execution that was creating the AIP
	 * - AIPCreator.deeplinkPatternCollection	Base URL for objects of type "collection"
	 * - AIPCreator.deeplinkPatternIE			Base URL for objects of type "intellectual entity" (IE)
	 * - AIPCreator.dbHost						Hostname/IP of the database server (including the JDBC prefix, e.g. "jdbc:mysql://localhost")
	 * - AIPCreator.dbPort						Port of the database server
	 * - AIPCreator.dbService					Database service to connect to
	 * - AIPCreator.dbUserName					Username for the database connection
	 * - AIPCreator.dbUserPassword				Password for the database connection
	 * 
	 * Optional properties for generateAIP()/confirmStorage():
	 * - AIPCreator.dbName						Database name to connect to
	 * 
	 * Mandatory properties for submit()/checkSubmission():
	 * - AIPCreator.ingestSubmitDir.[OS]		Target path where SIPs should be deposited for subsequent processing
	 * 
	 * Optional properties for submit()/checkSubmission():
	 * - AIPCreator.sshHost						Hostname/IP if SSH should be used
	 * - AIPCreator.sshPort						Port if SSH should be used, defaults to 22
	 * - AIPCreator.sshUser						Username if SSH should be used, defaults to current user
	 * - AIPCreator.sshPathToPK					Path to the private key if SSH should be used, defaults to ~/.ssh/id_rsa
	 * - AIPCreator.sshPathToKnownHosts			Path to the known hosts file if SSH should be used, defaults to ~/.ssh/known_hosts
	 */
	/* (non-Javadoc)
	 * @see ch.docuteam.darc.ingest.AIPCreator#initialize(java.util.Properties)
	 */
	@Override
	public void initialize(Properties properties) throws java.lang.Exception
	{
		// used for submit()/checkSubmission()
		this.ingestSubmitDir = properties.getProperty("AIPCreator.ingestSubmitDir." + OSSuffix);

		//	optional SSH properties
		this.sshHost = properties.getProperty("AIPCreator.sshHost");
		String sshPortString = properties.getProperty("AIPCreator.sshPort");
		if (sshPortString == null)
			this.sshPort = 22;
		else
			this.sshPort = Integer.parseInt(sshPortString);
		this.sshUser = properties.getProperty("AIPCreator.sshUser");
		if (this.sshUser == null)		this.sshUser = OperatingSystem.userName();
		this.sshPathToPK = properties.getProperty("AIPCreator.sshPathToPK");
		if (this.sshPathToPK == null)	this.sshPathToPK = OperatingSystem.userHome() + "/.ssh/id_rsa";
		this.sshPathToKnownHosts = properties.getProperty("AIPCreator.sshPathToKnownHosts");
		if (this.sshPathToKnownHosts == null)	this.sshPathToKnownHosts = OperatingSystem.userHome() + "/.ssh/known_hosts";

		// used for generateAIP()/confirmStorage()
		this.aipFolderPath = properties.getProperty("AIPCreator.aipFolderPath");

		this.executionId = properties.getProperty("AIPCreator.executionId");

		this.deeplinkPatternCollection = properties.getProperty("AIPCreator.deeplinkPatternCollection");
		this.deeplinkPatternIE = properties.getProperty("AIPCreator.deeplinkPatternIE");

		if (IsWithoutDB)					return;

		this.dbHost = properties.getProperty("AIPCreator.dbHost");
		this.dbPort = properties.getProperty("AIPCreator.dbPort");
		this.dbService = properties.getProperty("AIPCreator.dbService");
		// can be null, in which case the default database schema for the given user will be used
		this.dbName = properties.getProperty("AIPCreator.dbName", properties.getProperty("AIPCreator.dbUserName"));
		this.dbUserName = properties.getProperty("AIPCreator.dbUserName");
		this.dbUserPassword = properties.getProperty("AIPCreator.dbUserPassword");
	}


	/* (non-Javadoc)
	 * @see ch.docuteam.darc.ingest.AIPCreator#checkSubmission(ch.docuteam.darc.mets.Document, java.lang.String)
	 */
	@Override
	public List<String> checkSubmission(Document document)
	{
		checkPropertiesForSubmit();

		List<String> rejectMessages = new Vector<String>();
		String sipName = document.getSIPName();

		Logger.info("Performing submit-check in SIP: '" + sipName + "'");

		//	Is this document read-only? If yes, reject:
		if (document.isReadOnly())
		{
			Logger.debug("Can't submit because SIP is read-only: '" + sipName + "'");
			rejectMessages.add("MessageSubmitSIPIsReadOnly '" + sipName + "'");
		}

		//	Is this document locked? If yes, reject:
		if (document.isLocked())
		{
			Logger.debug("Can't submit because SIP is locked: '" + sipName + "'");
			rejectMessages.add("MessageSubmitSIPIsLocked '" + sipName + "'");
		}

		//	Is there at least one node with submitRequested? If no, reject:
		boolean hasNodesThatCanBeSubmitted = false;
		for (NodeAbstract n: document.getStructureMap().getRoot().getWithDescendants())
			if (n.getSubmitStatus() == SubmitStatus.SubmitRequested)
			{
				hasNodesThatCanBeSubmitted = true;
				break;
			}
		if (!hasNodesThatCanBeSubmitted)
		{
			Logger.debug("Can't submit because SIP has no submit requests: '" + sipName + "'");
			rejectMessages.add("MessageSubmitSIPHasNoSubmitRequests '" + sipName + "'");
		}

		//	Check each node in tree:
		Logger.debug("Checking all nodes in SIP: '" + sipName + "'");
		for (String message: this.submitCheckRecursive(document.getStructureMap().getRoot()))		rejectMessages.add(message);

		if (sshHost == null) {
			String submitSIPName = getSIPNameForSubmit(document);

			//	Check if this document already exists in the submit folder. If yes, reject:
			if (new File(this.ingestSubmitDir + File.separator + submitSIPName).exists())
			{
				rejectMessages.add("MessageSubmitSIPExistsAlreadyInSubmitFolder '" + submitSIPName + "'");
			}
		} else {
			SFTPChannel channel = new SFTPChannel(sshUser, sshHost, sshPort, sshPathToPK, sshPathToKnownHosts);
			try {
				channel.readFile("doesnotexist", ".");
			} catch (FileNotFoundException | SftpException | JSchException e) {
				//	ok, we don't expect to find an actual file there
				if (!(e instanceof SftpException && e.getMessage().equals("No such file")))
					rejectMessages.add("MessageSubmitSSHConnectionError " + sshUser + "@" + sshHost + ":" + sshPort + " with " + sshPathToPK + " and " + sshPathToKnownHosts);
			} finally {
				channel.closeConnection();
			}
		}

		if (rejectMessages.isEmpty())		Logger.debug("SIP is OK");
		else								Logger.debug("Can't submit SIP because: " + rejectMessages);

		return rejectMessages;
	}


	/**
	 * First check if submit can be issued.
	 * If yes, set all submit status entries that are "SubmitRequested" to "SubmitRequestPending", save and lock this document,
	 * then trigger the feeder-workflow "Ingest" (Here: by zipping the SIP into the submit folder).
	 * @return a list of error messages. If this list is empty, the submit was successful.
	 * @throws java.lang.Exception
	 */
	public List<String> submit(Document sip) throws java.lang.Exception
	{
		checkPropertiesForSubmit();

		//	Check document before submitting. If there are messages, don't submit but return the messages:
		List<String> rejectMessages = this.checkSubmission(sip);
		if (!rejectMessages.isEmpty())		return rejectMessages;
	
		String sipName = sip.getSIPName();
	
		//	Set SubmitStatus for parent entries of selected nodes to SubmitRequested so they get re-submitted
		for (NodeAbstract node: ((NodeFolder)sip.getStructureMap().getRoot()).getDescendants()) {
			if (node.getSubmitStatus().equals(SubmitStatus.SubmitRequested)) {
				NodeFolder nodeParent = (NodeFolder) node.getParent();
				while (nodeParent.getSubmitStatus().equals(SubmitStatus.Submitted)) {
					nodeParent.setSubmitStatus_force(SubmitStatus.SubmitRequested);
					
					if (nodeParent.isRoot())	break;
					
					nodeParent = (NodeFolder) nodeParent.getParent();
				}
			}
		}
	
		//	Set all selected nodes to "SubmitRequestPending":
		Logger.getLogger().debug("Setting all 'SubmitRequested' status entries to 'SubmitRequestPending' in SIP: '" + sipName + "'");
		sip.getStructureMap().getRoot().setSubmitStatusAllOrNone(SubmitStatus.SubmitRequested, SubmitStatus.SubmitRequestPending);
	
		//	Create a backup:
		sip.saveWithBackup();
	
		//	Remember the lock mode and file paths - in case I have to revert to the just created backup:
		Mode rememberedMode = sip.getMode();
		String rememberedFilePath = sip.getFilePath();
		String rememberedOriginalSIPFolder = sip.getOriginalSIPFolder();
		boolean rememberedIsWorkingCopy = sip.isWorkingCopy;
	
		//	Lock me as "SubmitRequestPending":
		Logger.getLogger().debug("Creating 'SubmitRequestPending' lock for SIP: '" + sipName + "'");
		sip.lockAsSubmitRequestPending();
	
		//	For not-zipped SIPs, remove the working copy and switch to the original SIP as read-only; for ZIP-files, the working copy will be removed anyway when closing this SIP:
		if (!sip.isZIPFile())
		{
			//	Remove the working copy:
			sip.cleanupWorkingCopy();
	
			// 	Switch to the original SIP:
			sip.setFilePath(FileUtil.asCanonicalFileName(sip.getOriginalSIPFolder() + File.separator + Document.DEFAULT_METS_FILE_NAME));
			sip.isWorkingCopy = false;
		}
		
		//	Create the list of files that the submitted SIP should contain (we don't have to include nodes that are not to be submitted)
		List<String> sourcePaths = new ArrayList<String>();
		sourcePaths.add(sip.getSIPFolder() + "/mets.xml");
		for (NodeAbstract node: sip.getStructureMap().getRoot().getWithDescendants()) {
			//	TODO	Enable submission of folders without files, either by creating a temporary copy of the structure or by modifying the Zipper class
			if (node.getSubmitStatus().equals(SubmitStatus.SubmitRequestPending) && node.isFile())
				sourcePaths.add(node.getAbsolutePathString());
		}
	
		try
		{
			//	Trigger feeder-workflow "Ingest" by placing a ZIP of this SIP into the ingest folder:
			Logger.getLogger().debug("Zipping to ingest-folder: '" + ingestSubmitDir + "' SIP: '" + sipName + "'");
			if (sshHost == null) {
				Zipper.zip(sourcePaths.toArray(new String[]{}), ingestSubmitDir + File.separator + this.getSIPNameForSubmit(sip));
				//	Do also push the currently used levels.xml to the submit location (matching the location expected by the feeder operation)
				FileUtil.copyToOverwriting(LevelOfDescription.getInitializationFilePath(), new File(ingestSubmitDir).getParent() + "/config/levels.xml");
			} else {
				File tempZIPFile = new File(FileUtil.getTempFolder() + File.separator + UniqueID.getString() + File.separator + getSIPNameForSubmit(sip));
				tempZIPFile.getParentFile().mkdirs();
				FileUtil.deleteOnExit(tempZIPFile.getParentFile());
				Zipper.zip(sourcePaths.toArray(new String[]{}), tempZIPFile.getAbsolutePath());
				Logger.getLogger().debug("Using SFTP: " + sshUser + "@" + sshHost + ":" + sshPort + " with " + sshPathToPK + " and " + sshPathToKnownHosts);
				SFTPChannel channel = new SFTPChannel(sshUser, sshHost, sshPort, sshPathToPK, sshPathToKnownHosts);
				channel.transferFileLeaveSessionOpen(tempZIPFile.getAbsolutePath(), getSIPNameForSubmit(sip), ingestSubmitDir);
				//	Do also push the currently used levels.xml to the submit location (matching the location expected by the feeder operation)
				channel.transferFileLeaveSessionOpen(new File(LevelOfDescription.getInitializationFilePath()).getAbsolutePath(), "levels.xml", "../config");
			}
		}
		catch (java.lang.Exception e)
		{
			Logger.getLogger().error("Copying to submit folder failed, reverting to original SIP: '" + sipName + "'", e);
			
			//	Prepare for reverting to last backup: remove the SubmitRequestPending lock, reset the original lock and the file paths:
			sip.unlock_dontCheck();
			sip.setMode(rememberedMode);
			sip.setFilePath(rememberedFilePath);
			sip.setOriginalSIPFolder(rememberedOriginalSIPFolder);
			sip.isWorkingCopy = rememberedIsWorkingCopy;
			sip.revertToLastBackup();
			
			throw new java.lang.Exception("Copying to submit folder failed, reverting to original SIP: '" + sipName + "'", e);
		}
	
		return rejectMessages;
	}


	/**
	 * Test if the folder ingestFeedbackDir contains a folder that exists in the workspaceFolderPath,
	 * and that this folder contains only and exactly the file "mets.xml".
	 * If yes, copy the file "mets.xml" and overwrite it in the SIP in the workspace, then delete it from the ingestFeedackFolder
	 * and the zipped SIP from the ingestSubmitFolder.
	 * This works for zipped SIPs as well.
	 * @param workspaceFolderPath
	 * @throws FileUtilExceptionListException
	 * @throws IOException
	 */
	public List<String> checkIngestFeedback(String workspaceFolderPath) throws IOException, FileUtilExceptionListException
	{
		checkPropertiesForSubmit();

		String ingestFeedbackDir = null;
		List<String> ingestFeedbackFoundForSIPs = new Vector<String>();
		SFTPChannel channelSftp = null;
		ChannelSftp basicChannelSftp = null;

		Logger.getLogger().debug("Checking ingest feedback for workspace folder: '" + workspaceFolderPath + "'");

		if (workspaceFolderPath == null)						return ingestFeedbackFoundForSIPs;
		File workspaceFolder = new File(workspaceFolderPath);
		if (!workspaceFolder.exists())							return ingestFeedbackFoundForSIPs;
		if (!workspaceFolder.isDirectory())						return ingestFeedbackFoundForSIPs;
		if (workspaceFolder.list().length == 0)					return ingestFeedbackFoundForSIPs;

		// Check for feedback
		List<String> feedbackFilenames = new ArrayList<String>();
		if (sshHost == null) {
			ingestFeedbackDir = new File(this.ingestSubmitDir).getParent() + FeedbackRelativeFolderName;

			// use local file paths
			File ingestFeedbackFolder = new File(ingestFeedbackDir);
			if (!ingestFeedbackFolder.exists())						return ingestFeedbackFoundForSIPs;
			if (!ingestFeedbackFolder.isDirectory())				return ingestFeedbackFoundForSIPs;

			feedbackFilenames = Arrays.asList(ingestFeedbackFolder.list((FilenameFilter)FileFilter.SIPs));
		} else {
			ingestFeedbackDir = this.ingestSubmitDir + "/.." + FeedbackRelativeFolderName;

			// use SSH connection
			Logger.getLogger().debug("Using SSH: " + sshUser + "@" + sshHost + ":" + sshPort + " with " + sshPathToPK + " and " + sshPathToKnownHosts);
			channelSftp = new SFTPChannel(sshUser, sshHost, sshPort, sshPathToPK, sshPathToKnownHosts);
			try {
				basicChannelSftp = channelSftp.getChannelSftp();
			} catch (FileNotFoundException | SftpException | JSchException e) {
				Logger.getLogger().error("Error opening SFTP channel", e);
				return ingestFeedbackFoundForSIPs;
			}
			
			Vector<LsEntry> feedbackEntries;
			try {
				feedbackEntries = basicChannelSftp.ls(ingestFeedbackDir);
			} catch (SftpException e) {
				Logger.getLogger().error("Error listing the feedback folder content on the SFTP host", e);
				return ingestFeedbackFoundForSIPs;
			}
			try {
				for (LsEntry entry : feedbackEntries) {
					Vector<LsEntry> subentries = basicChannelSftp.ls(ingestFeedbackDir + "/" + entry.getFilename());
					for (LsEntry subentry : subentries) {
						if (subentry.getFilename().equalsIgnoreCase("mets.xml")) {
							feedbackFilenames.add(entry.getFilename());
						}
					}
				}
			} catch (SftpException e) {
				Logger.getLogger().error("Error listing individual feedback folders", e);
				return ingestFeedbackFoundForSIPs;
			}
		}
		if (feedbackFilenames.size() == 0)		return ingestFeedbackFoundForSIPs;

		for (String feedbackSIPName: feedbackFilenames)
		{
			Logger.getLogger().debug("Found ingest feedback for SIP: '" + feedbackSIPName + "'");

			String sipName = null;
			String sipPath = null;
			String feedbackSIPPath = null;
			
			if (feedbackSIPName.endsWith(UNDERSCORE_ZIP))
			{
				//	SIP is zipped:
	
				//	Replace the trailing "_zip" by ".zip":
				sipName = feedbackSIPName.substring(0, feedbackSIPName.length() - 4) + ZIP_EXT;
				sipPath = workspaceFolder + "/" + sipName;
	
				if (!new File(sipPath).exists())
				{
					Logger.getLogger().debug("No SIP in Workspace: '" + sipName + "'");
					continue;
				}
				if (!new File(sipPath).isFile())
				{
					Logger.getLogger().debug("SIP is not a file: '" + sipName + "'");
					continue;
				}

				//	If this document is not locked as SubmitRequestPending, ignore it:
				String lockedBy = Document.lockedByWhom(sipPath);
				if (!	(Document.isLocked(sipPath)
						&& Document.lockedByWhom(sipPath).contains(SubmitStatus.SubmitRequestPending.toString())))
				{
					Logger.getLogger().debug("No SubmitRequestPending lock: '" + lockedBy + "' for SIP: '" + sipName + "'");
					continue;
				}
	
				Logger.getLogger().debug("Replacing METS-file from ingest feedback for SIP: '" + sipName + "'");
	
				feedbackSIPPath = ingestFeedbackDir + "/" + feedbackSIPName;

				// Replace mets.xml in zipped SIP
				File feedbackMETSFile;
				if (sshHost == null) {
					feedbackMETSFile = new File(feedbackSIPPath + "/mets.xml");
				} else {
					feedbackMETSFile = new File(FileUtil.getTempFolder() + "mets" + UniqueID.getXML() + ".xml");
					feedbackMETSFile.deleteOnExit();
					try {
						basicChannelSftp.get(feedbackSIPPath + "/" + Document.DEFAULT_METS_FILE_NAME, feedbackMETSFile.getAbsolutePath());
					} catch (SftpException e) {
						Logger.getLogger().debug("Could not fetch updated mets.xml for: '" + feedbackSIPName + "'", e);
						continue;
					}
				}
				TFile.cp(feedbackMETSFile, new TFile(sipPath, Document.DEFAULT_METS_FILE_NAME));
				TFile.umount();
			}
			else
			{
				//	SIP is folder:
				sipPath = workspaceFolder + "/" + feedbackSIPName;
	
				if (!new File(sipPath).exists())
				{
					Logger.getLogger().debug("No SIP in Workspace: '" + feedbackSIPName + "'");
					continue;
				}
				if (!new File(sipPath).isDirectory())
				{
					Logger.getLogger().debug("SIP is not a directory: '" + feedbackSIPName + "'");
					continue;
				}
	
				//	If this document is not locked as SubmitRequestPending, ignore it:
				String lockedBy = Document.lockedByWhom(sipPath);
				if (!	(Document.isLocked(sipPath)
						&& lockedBy.contains(SubmitStatus.SubmitRequestPending.toString())))
				{
					Logger.getLogger().debug("No SubmitRequestPending lock: '" + lockedBy + "' for SIP: '" + feedbackSIPName + "'");
					continue;
				}
	
				Logger.getLogger().debug("Replacing METS-file from ingest feedback for SIP: '" + feedbackSIPName + "'");
	
				feedbackSIPPath = ingestFeedbackDir + "/" + feedbackSIPName;
	
				//	Replace mets-file:
				File feedbackMETSFile;
				if (sshHost == null) {
					feedbackMETSFile = new File(feedbackSIPPath + "/mets.xml");
				} else {
					feedbackMETSFile = new File(FileUtil.getTempFolder() + "/mets" + UniqueID.getXML() + ".xml");
					feedbackMETSFile.deleteOnExit();
					try {
						basicChannelSftp.get(feedbackSIPPath + "/" + Document.DEFAULT_METS_FILE_NAME, feedbackMETSFile.getAbsolutePath());
					} catch (SftpException e) {
						Logger.getLogger().debug("Could not fetch updated mets.xml for: '" + feedbackSIPName + "'", e);
						continue;
					}
				}
				FileUtil.copyToOverwriting(feedbackMETSFile.getAbsolutePath(), sipPath + "/" + Document.DEFAULT_METS_FILE_NAME);
			}

			//	Unlock:
			FileUtil.delete(sipPath + Document.LOCK_FILE_SUFFIX);

			String ingestSIPPath = ingestSubmitDir + "/" + feedbackSIPName + ZIP_EXT;
			if (sshHost == null) {
				//	Cleanup feedback folder:
				FileUtil.delete(feedbackSIPPath);
	
				//	Cleanup ingest folder:
				FileUtil.delete(ingestSIPPath);
			} else {
				//	Cleanup feedback folder:
				try {
					basicChannelSftp.rm(feedbackSIPPath + "/" + Document.DEFAULT_METS_FILE_NAME);
					basicChannelSftp.rmdir(feedbackSIPPath);
				} catch (SftpException e) {
					Logger.getLogger().error("Error cleaning feedback directory: '" + feedbackSIPPath + "'", e);
					continue;
				}

				//	Cleanup ingest folder:
				try {
					basicChannelSftp.rm(ingestSIPPath);
				} catch (SftpException e) {
					Logger.getLogger().error("Error cleaning ingest directory: '" + ingestSIPPath + "'", e);
					continue;
				}
			}

			ingestFeedbackFoundForSIPs.add(feedbackSIPName);
		}
	
		return ingestFeedbackFoundForSIPs;
	}


	/**
	 * Return the SIP fileName to be used when submitting this SIP. Since the submitted SIP will be zipped, it must have the ".zip" suffix.
	 * In addition, if the *original* SIP is zipped, the file name contains "_zip" for distinction; this is needed for the ingest feedback.
	 * @return
	 */
	public String getSIPNameForSubmit(Document sip)
	{
		return (sip.isZIPFile() ? FileUtil.asFileNameWithoutExtension(sip.getSIPName()) + "_zip" : sip.getSIPName()) + ".zip";
	}


	/**
	 * Return the SIP fileName to be used when submitting this SIP. Since the submitted SIP will be zipped, it must have the ".zip" suffix.
	 * In addition, if the *original* SIP is zipped, the file name contains "_zip" for distinction; this is needed for the ingest feedback.
	 * @return
	 */
	public String getSIPNameWithoutExtension(Document sip)
	{
		return sip.isZIPFile() ? FileUtil.asFileNameWithoutExtension(sip.getSIPName()) : sip.getSIPName();
	}


	/* (non-Javadoc)
	 * @see ch.docuteam.darc.ingest.AIPCreator#generateAIP(ch.docuteam.darc.mets.Document)
	 */
	@Override
	public List<String> generateAIP(Document document) throws java.lang.Exception
	{
		checkPropertiesForStorage();

		//	Cut away the ".zip" from the SIP name in case we're working directly on a zipped SIP:
		String sipName = getSIPNameWithoutExtension(document);

		List<String> rejectMessages = new Vector<String>();

		if (this.aipFolderPath == null)
		{
			Logger.debug("Can't submit because AIPFolderPath is undefined: '" + sipName + "'");
			rejectMessages.add("MessageSubmitAIPUndefinedSubmitFolderPath '" + sipName + "'");

			//	Can't continue, so return:
			return rejectMessages;
		}

		String workspace = document.getStructureMap().getRoot().getDynamicMetadataValueForName("refCodeAdmin");
		String aipPath = this.aipFolderPath + File.separator + workspace + File.separator + sipName;

		if (IsTestRun)			this._prepareTestRun(aipPath);

		//	Does this AIP already exist in the submit folder? If yes, reject:
		if (new File(aipPath).exists())
		{
			Logger.debug("Can't submit because AIP: '" + aipPath + "' exists already.");
			rejectMessages.add("MessageSubmitAIPExistsAlreadyInSubmitFolder '" + aipPath + "'");

			//	Can't continue, so return:
			return rejectMessages;
		}

		try
		{
			DB.initialize(this.dbHost, this.dbPort, this.dbService, this.dbName, this.dbUserName, this.dbUserPassword, this.executionId);

			for (NodeAbstract node: document.getStructureMap().getRoot().getWithDescendants()) {
				if (IsTestRun)		DB._prepareTestRun(workspace, node.getPathString(), sipName);
				//	Was any of the selected nodes of type IE already ingested, i.e. does it already exist in the DB? If yes, reject:
				if (node.getSubmitStatus().equals(SubmitStatus.SubmitRequestPending) && this.getAIPType(node).equals(AIPType.IE) && DB.wasAlreadyIngested(workspace, sipName, node))
				{
					Logger.debug("Can't submit because AIP: '" + aipPath + "/" + node.getPathString() + "' was already ingested.");
					rejectMessages.add("MessageSubmitAIPWasAlreadyIngested '" + aipPath + "/" + node.getPathString() + "'");
				}
			}
			if (!rejectMessages.isEmpty()) {
				//	Can't continue, so return:
				return rejectMessages;
			}

			FileUtil.createFolderOverwriting(aipPath);

			String submitTimestamp = document.getHeader().getLastModificationDate().replace("T", " ");

			rejectMessages.addAll(this.generateAIP(document.getStructureMap().getRoot(), workspace, sipName, submitTimestamp));
			if (rejectMessages.isEmpty())	DB.logAIPCreationFinalizeOK(workspace, sipName);
		}
		finally
		{
			DB.close();
		}

		return rejectMessages;
	}


	/* (non-Javadoc)
	 * @see ch.docuteam.darc.ingest.AIPCreator#confirmStorage(ch.docuteam.darc.mets.Document)
	 */
	/**
	 * If ...
	 */
	@Override
	public void confirmStorage(Document document) throws SQLException, IllegalArgumentException, IllegalAccessException, InvocationTargetException, MetadataElementIsNotDefinedException, MetadataElementIsNotAllowedException
	{
		checkPropertiesForStorage();

		String sipName = getSIPNameWithoutExtension(document);
		String workspace = document.getStructureMap().getRoot().getDynamicMetadataValueForName("refCodeAdmin");

		Logger.debug("Checking storage status of SIP: '" + sipName + "'");

		Map<String, IngestStatus> mapIngestStatus = new HashMap<String, IngestStatus>();
		Map<String, String> mapEntityIDs = new HashMap<String, String>();
		try
		{
			DB.initialize(this.dbHost, this.dbPort, this.dbService, this.dbName, this.dbUserName, this.dbUserPassword, this.executionId);
			mapIngestStatus = DB.getAllStatus(workspace, sipName);
			mapEntityIDs = DB.getAllEntityIDs(workspace, sipName);
		}
		finally
		{
			DB.close();
		}

		if (mapIngestStatus.isEmpty() || mapEntityIDs.isEmpty())
		{
			Logger.info("Nothing found in DB for SIP: '" + sipName + "'");
			return;
		}

		//	If at least one status is neither FINISHED nor REJECTED nor DECLINED, this AIP is not yet finished:
		boolean isFinished = true;
		for (String amdId: mapIngestStatus.keySet())
		{
			IngestStatus ingestStatus = mapIngestStatus.get(amdId);

			Logger.debug("For SIP: '" + sipName + "' id: '" + amdId + "' found status: '" + ingestStatus + "'");

			if (!(	ingestStatus == IngestStatus.FINISHED
				 || ingestStatus == IngestStatus.REJECTED
				 || ingestStatus == IngestStatus.DECLINED
				) )
			{
				isFinished = false;
				break;
			}
		}

		//	Ignore if not finished:
		if (!isFinished)
		{
			Logger.info("Not yet finished: '" + sipName + "'");
			return;
		}

		//	Finished: now go through the document and set the status for all nodes:
		boolean exceptionOccurred = false;
		for (String amdId: mapIngestStatus.keySet())
		{
			NodeAbstract node = document.getStructureMap().searchId(amdId);
			if (node != null)
			{
				try
				{
					//	write the Rosetta deeplink into the respective metadata field
					switch (this.getAIPType(node)) {
						case IE:
							node.setDynamicMetadataValueForName_force("locationOfOriginals", String.format(this.deeplinkPatternIE, mapEntityIDs.get(amdId)));
							break;
						case Collection:
							node.setDynamicMetadataValueForName_force("locationOfOriginals", String.format(this.deeplinkPatternCollection, mapEntityIDs.get(amdId)));
							break;
						case None:
						default:
							break;
					}
					
					IngestStatus ingestStatus = mapIngestStatus.get(amdId);
					switch (ingestStatus)
					{
						case FINISHED:
						{
							node.setSubmitStatus(SubmitStatus.Submitted);

							//	If this node is a complex IE with multiple subfiles, set all children's submit status too:
							if (node.isFolder() && this.getAIPType(node) == AIPType.IE)
								for (MutableTreeTableNode child: ((NodeFolder)node).getDescendants())
									((NodeAbstract)child).setSubmitStatus(SubmitStatus.Submitted);

							break;
						}
						case REJECTED:
						case DECLINED:
						{
							node.setSubmitStatus(SubmitStatus.SubmitFailed);

							//	If this node is a complex IE with multiple subfiles, set all children's submit status too:
							if (node.isFolder() && this.getAIPType(node) == AIPType.IE)
								for (MutableTreeTableNode child: ((NodeFolder)node).getDescendants())	((NodeAbstract)child).setSubmitStatus(SubmitStatus.SubmitFailed);

							break;
						}
						default:
							break;
					}
				} catch (CantSetSubmitStatusNotAllowedException e) {
					e.printStackTrace();
					exceptionOccurred = true;
				} catch (MetadataElementValidatorException e) {
					e.printStackTrace();
					exceptionOccurred = true;
				} catch (LevelMetadataElementIsReadOnly e) {
					e.printStackTrace();
					exceptionOccurred = true;
				} catch (MetadataElementSetterPostActionException e) {
					e.printStackTrace();
					exceptionOccurred = true;
				} catch (MetadataElementAllowedValuesException e) {
					e.printStackTrace();
					exceptionOccurred = true;
				}
			}
		}

		if (exceptionOccurred)
		{
			//	TODO: What to do when an exception occurred?
		}
		else
		{
			//	Copy the mets.xml file to the feedback folder, which is the folder named "feedback" next to the dropbox folder:
			try
			{
				String feedbackPath = new File(document.getOriginalSIPFolder()).getParentFile().getParent() + FeedbackRelativeFolderName + File.separator + sipName;

				Logger.debug("Sending feedback for SIP: '" + sipName + "' to folder: '" + feedbackPath + "'");

				document.saveWithoutBackup();
				document.unlockIfNecessary();
				document.cleanupWorkingCopy();

				FileUtil.copyToFolderOverwriting(document.getFilePath(), feedbackPath);

				//	Delete the SIP from the drop-folder:
				FileUtil.delete(document.getOriginalSIPFolder());

				//	TODO	Send email
			}
			catch (Exception ex)
			{
				//	TODO	What to do when an exception occurs?
				ex.printStackTrace();
			}
		}
	}

	//	--------		Actions				-------------------------------------------------------
	//	--------		Business Ops		-------------------------------------------------------
	//	--------		Persistence			-------------------------------------------------------
	//	--------		Support				-------------------------------------------------------
	//	--------		Utilities			-------------------------------------------------------
	//	---------		Misc				-------------------------------------------------------
	//	--------		Debugging			-------------------------------------------------------
	//	---------		Temporary			-------------------------------------------------------

	//	ToDo: This is temporary during development:
	public Map<String, IngestStatus> _getAllStatus(Document document) throws IllegalArgumentException, IllegalAccessException, InvocationTargetException, MetadataElementIsNotAllowedException, MetadataElementIsNotDefinedException, SQLException {
		//	Remove the trailing ".zip" from the SIP name:
		String sipName = getSIPNameWithoutExtension(document);
		String workspace = document.getStructureMap().getRoot().getDynamicMetadataValueForName("refCodeAdmin");

		Logger.debug("Getting all status entries for SIP: '" + sipName + "'");

		try
		{
			DB.initialize(this.dbHost, this.dbPort, this.dbService, this.dbName, this.dbUserName, this.dbUserPassword, this.executionId);
			return DB.getAllStatus(workspace, sipName);
		}
		finally
		{
			DB.close();
		}
	}
	
	public void _setAllStatus(Document document, IngestStatus status) throws IllegalArgumentException, IllegalAccessException, InvocationTargetException, MetadataElementIsNotDefinedException, MetadataElementIsNotAllowedException, SQLException
	{
		//	Remove the trailing ".zip" from the SIP name:
		String sipName = getSIPNameWithoutExtension(document);
		String workspace = document.getStructureMap().getRoot().getDynamicMetadataValueForName("refCodeAdmin");

		Logger.debug("Setting all status entries for SIP: '" + sipName + "' to: '" + status + "'");

		try
		{
			DB.initialize(this.dbHost, this.dbPort, this.dbService, this.dbName, this.dbUserName, this.dbUserPassword, this.executionId);
			DB._setAllStatus(status, workspace, sipName);
		}
		finally
		{
			DB.close();
		}
	}

	//	========	Instance Private		=======================================================

	//	--------		Initializing		-------------------------------------------------------

	/**
	 * Check that the necessary properties are available to do any submissions.
	 */
	private void checkPropertiesForSubmit() {
		if (this.ingestSubmitDir == null)	throw new NullPointerException("Missing property: 'AIPCreator.ingestSubmitDir." + OSSuffix + "'");
	}
	
	/**
	 * Check that the necessary properties are available to do any storage operations.
	 */
	private void checkPropertiesForStorage() {
		if (this.aipFolderPath == null)		throw new NullPointerException("Missing property: 'AIPCreator.aipFolderPath'");
		if (this.executionId == null)		throw new NullPointerException("Missing property: 'AIPCreator.executionId'");
		if (this.deeplinkPatternCollection == null)	throw new NullPointerException("Missing property: 'AIPCreator.deeplinkPatternCollection'");
		if (this.deeplinkPatternIE == null)	throw new NullPointerException("Missing property: 'AIPCreator.deeplinkPatternIE'");

		if (IsWithoutDB)					return;
		if (this.dbHost == null)			throw new NullPointerException("Missing property: 'AIPCreator.dbHost'");
		if (this.dbPort == null)			throw new NullPointerException("Missing property: 'AIPCreator.dbPort'");
		if (this.dbService == null)			throw new NullPointerException("Missing property: 'AIPCreator.dbService'");
		if (this.dbUserName == null)		throw new NullPointerException("Missing property: 'AIPCreator.dbUserName'");
		if (this.dbUserPassword == null)	throw new NullPointerException("Missing property: 'AIPCreator.dbUserPassword'");
	}
	
	//	--------		Accessing			-------------------------------------------------------
	//	--------		Inquiring			-------------------------------------------------------

	private List<String> submitCheckRecursive(NodeAbstract node)
	{
		List<String> rejectMessages = new Vector<String>();

		//	Check each subnode if it could be submitted:
		for (NodeAbstract n: node.getWithDescendants())
			if (n.getSubmitStatus().equals(SubmitStatus.SubmitRequested))
				for (String message: this.submitCheck(n))
					rejectMessages.add(message);

		return rejectMessages;
	}


	private List<String> submitCheck(NodeAbstract node)
	{
		List<String> rejectMessages = new Vector<String>();
		String pathString = node.getPathString();

		Logger.info("Performing submit-check for node: '" + pathString + "' (" + node.getAdmId() + ")");

		if (!node.fileExists())											rejectMessages.add("MessageSubmitNodeFileDoesNotExist '" + pathString + "'");
		if (!node.canRead())											rejectMessages.add("MessageSubmitNodeFileIsNotReadable '" + pathString + "'");
		if (!node.canWrite())											rejectMessages.add("MessageSubmitNodeFileIsNotWritable '" + pathString + "'");

		//	Is SubmitStatus = "SubmitRequestPending"? If yes, reject (SIP is currently in an ongoing submission process):
		//	TODO	does not make sense anymore
		SubmitStatus submitStatus = node.getSubmitStatus();
		if (submitStatus == SubmitStatus.SubmitRequestPending)			rejectMessages.add("MessageSubmitNodeIsAlreadySubmitPending '" + pathString + "'");

		//	Are any mandatory MD fields not set? If yes, reject:
		if (node.hasDynamicMetadataElementInstancesWhichAreMandatoryButNotSet())
																		rejectMessages.add("MessageSubmitNodeMandatoryMetadataNotSet '" + pathString + "'");
		
		//	Test if all children have either the level "undefined" or not:
		Boolean allChildrenLevelsAreUndef = null;
		for (int i = 0; i < node.getChildCount(); i++)
		{
			boolean childLevelIsUndef = ((NodeAbstract)node.getChildAt(i)).getLevel().isUndefined();
			if (allChildrenLevelsAreUndef == null)
				allChildrenLevelsAreUndef = childLevelIsUndef;
			else
				if (allChildrenLevelsAreUndef != childLevelIsUndef)
				{
					rejectMessages.add("MessageSubmitNodeHasHeterogenousChildren '" + pathString + "'");
					break;
				}
		}

		if (rejectMessages.isEmpty())		Logger.info("Node is OK");
		else								Logger.info("Can't submit node because: " + rejectMessages);

		return rejectMessages;
	}


	private List<String> generateAIP(NodeAbstract node, String workspace, String sipName, String submitTimestamp) throws SQLException
	{
		List<String> errorMessages = new Vector<String>();

		switch (this.getAIPType(node))
		{
			case IE:
			{
				errorMessages = this.generateIE(node, workspace, sipName, submitTimestamp);
				break;
			}
			case Collection:
			{
				errorMessages = this.generateCollection((NodeFolder)node, workspace, sipName, submitTimestamp);

				//	Recurse into the children, collect possible error messages:
				for (int i = 0; i < node.getChildCount(); i++)
					errorMessages.addAll(this.generateAIP((NodeAbstract)node.getChildAt(i), workspace, sipName, submitTimestamp));

				break;
			}
			case None:
				Logger.info("Skipping in '" + sipName + "' node: '" + node.getPathString() + "': Neither IE nor Collection");
				break;
			default:
				Logger.error("This should't ever happen!");
				break;
		}

		return errorMessages;
	}


	private AIPType getAIPType(NodeAbstract node)
	{
		//	If the node is a file and has the level "undefined", ignore it, otherwise create an IE:
		if (node.isFile())
			return node.getLevel().isUndefined()
				? AIPType.None
				: AIPType.IE;

		//	If this folder is empty, ignore it:
		if (node.getChildCount() == 0)		return AIPType.None;

		//	If the folder's level is undefined, create a collection and recurse in:
		if (node.getLevel().isUndefined())	return AIPType.Collection;

		//	Only if all node's children's level is undefined, create a complex IE with multiple subfiles.
		//	If at least one child's level is NOT undefined, create a Collection and recurse in:
		for (int i = 0; i < node.getChildCount(); i++)
			if (!((NodeAbstract)node.getChildAt(i)).getLevel().isUndefined())
				return AIPType.Collection;

		return AIPType.IE;
	}

	//	--------		Low-Level Ingest	-------------------------------------------------------

	/**
	 * Create a Collection in the submit path.
	 * @param submitPath
	 * @throws Exception
	 * @author denis
	 * @return Error message or empty string
	 * Don't do anything if the node's submit status is NOT SubmitStatusPending.
	 * @throws SQLException
	 */
	private List<String> generateCollection(NodeFolder node, String workspace, String sipName, String submitTimestamp) throws SQLException
	{
		List<String> errorMessages = new Vector<String>();

		if (node.getSubmitStatus() != SubmitStatus.SubmitRequestPending)
		{
			Logger.info("Skipping in '" + sipName + "' node: '" + node.getPathString() + "': Submit status is not '" + SubmitStatus.SubmitRequestPending + "'");
			return errorMessages;
		}

		String aipPath = this.aipFolderPath + File.separator + workspace + File.separator + sipName;
		String aipFilePath = aipPath + "/" + FileUtil.asSafeFileName(node.getLabel()) + node.getAdmId() + ".xml";

		Logger.info("Creating Collection in '" + aipPath + "' node: '" + node.getPathString() + "'");

		DB.logCollectionCreationStart(node, aipFilePath, workspace, sipName, submitTimestamp);

		try
		{
			//	Create new xml document with a root element:
			org.dom4j.Document document = DocumentHelper.createDocument();

			//	Fill it with data:
			Element collectionElement = document.addElement("Collection");
			collectionElement.addElement("url").addText(workspace + "/" + node.getPathString());
			collectionElement.addElement("externalSystem").addText(workspace);
			collectionElement.addElement("name").addText(node.getLabel());
			if (node.getPID() != null)		collectionElement.addElement("externalId").addText(node.getPID().get(0));

			this.addCollectionMetadataElementToCollectionElement(collectionElement, "dc:title", node.getUnitTitle().substring(0, Math.min(node.getUnitTitle().length(), 127)));
			if (node.isRoot())
				this.addCollectionMetadataElementToCollectionElement(collectionElement, "dcterms:isPartOf", workspace);
			else
				this.addCollectionMetadataElementToCollectionElement(collectionElement, "dcterms:isPartOf", workspace + "/" + ((NodeAbstract)node.getParent()).getPathString());

			for(String key: MetadataNamesMapToDC.keySet())
				this.addDynamicMetadataValuesToCollectionElement(node, workspace, key, collectionElement, MetadataNamesMapToDC.get(key));

			//	Save it:
			Writer oswriter = new java.io.OutputStreamWriter(new FileOutputStream(aipFilePath), "UTF-8");
			OutputFormat.createPrettyPrint();
			OutputFormat outformat = new OutputFormat("	", false, "UTF-8");
			XMLWriter writer = new XMLWriter(oswriter, outformat);
			writer.write(document);
			writer.close();
		}
		catch (java.lang.Exception ex)
		{
			ex.printStackTrace();
			errorMessages.add("MessageSubmitNodeExceptionCreatingCollection '" + aipPath + "': " + ex.getMessage());
		}

		DB.logCollectionCreationEnd(node, errorMessages.isEmpty()? IngestStatus.FEEDER_FINISHED: IngestStatus.FEEDER_ERROR, workspace, sipName);

		return errorMessages;
	}


	private void addDynamicMetadataValuesToCollectionElement(NodeAbstract node, String workspace, String mdName, Element collectionElement, String elementName)
	{
		Logger.debug("Mapping metadata: '" + mdName + "'");

		try
		{
			for (String s: node.getAllDynamicMetadataValuesForName(mdName))
				if (s != null && !s.isEmpty())
				{
					//	If there is a mapping in the DB Mapping Table for this mdName and mdValue, save the mapping, otherwise the mdValue itself:
					String value = DB.getRosettaCodeIfExists(workspace, mdName, s);
					this.addCollectionMetadataElementToCollectionElement(collectionElement, elementName, value);

					Logger.debug("Mapping: '" + mdName + "' to: '" + elementName + "' value: '" + value + "'");
				}
		}
		catch (Exception ex)
		{
			//	Ignore exception (e.g. when the dyn. metadata value is not accessable for this level) and just continue
		}
	}


	private void addCollectionMetadataElementToCollectionElement(Element collectionElement, String key, String value)
	{
		Element collectionMetaDataElement = collectionElement.addElement("CollectionMetaData");
		collectionMetaDataElement.addElement("key").addText(key);
		collectionMetaDataElement.addElement("value").addText(value);
	}


	//	TODO	split into several methods to keep it readable (and understandable)
	/**
	 * Create an IE out of the node passed. There are 2 different modes:
	 * 	1: node is a file or a single folder created by a migration event (= simple IE)
	 * 	2: node is a folder with children (= complex IE)
	 * Don't do anything if the node's submit status is NOT SubmitStatusPending.
	 * The name of the IE is the node's label (= the file name) as safe filename, followed by its admid.
	 * @param node
	 * @param workspace
	 * @param sipName
	 * @param submitTimestamp
	 * @return Error message or empty string
	 * @throws SQLException
	 */
	@SuppressWarnings("deprecation")
	private List<String> generateIE(NodeAbstract node, String workspace, String sipName, String submitTimestamp) throws SQLException
	{
		List<String> errorMessages = new Vector<String>();

		if (node.getSubmitStatus() != SubmitStatus.SubmitRequestPending)
		{
			Logger.info("Skipping in '" + sipName + "' node: '" + node.getPathString() + "': Submit status is not 'SubmitStatusPending'");
			return errorMessages;
		}

		String aipPath = this.aipFolderPath + File.separator + workspace + File.separator + sipName;
		String aipRootPath = aipPath + "/" + FileUtil.asSafeFileName(node.getLabel()) + node.getAdmId();
		String ieFilePath = aipRootPath + "/content/ie.xml";
		String dcFilePath = aipRootPath + "/content/dc.xml";
		String streamsFilePath = aipRootPath + "/content/streams";

		Logger.info("Creating DC in '" + aipPath + "' node: '" + node.getLabel() + "'");

		DB.logIECreationStart(node, aipRootPath, workspace, sipName, submitTimestamp);

		try
		{
			//	Create dc.xml:
			DublinCore dc = DublinCoreFactory.getInstance().createDocument();
			//	Take only first 127 characters of node's unitTitle:
			dc.addElement("dc:title", node.getUnitTitle().substring(0, Math.min(node.getUnitTitle().length(), 127)));
			new File(dcFilePath).getParentFile().mkdirs();
			dc.save(dcFilePath);

			//	Create ie.xml:
			Logger.info("Creating IE in '" + aipPath + "' node: '" + node.getLabel() + "'");

			IEParser ie = IEParserFactory.create();

			//	Add DC to IE:
			Logger.debug("Adding DC to IE in '" + aipPath + "' node: '" + node.getLabel() + "'");

			dc = ie.getDublinCoreParser();

			dc.addElement("dc:title", node.getUnitTitle().substring(0, Math.min(node.getUnitTitle().length(), 127)));
			dc.addElement("dcterms:isPartOf", workspace + "/" + ((NodeAbstract)node.getParent()).getPathString());

			for(String key: MetadataNamesMapToDC.keySet())
				this.addDynamicMetadataValuesToDCElement(node, workspace, key, dc, MetadataNamesMapToDC.get(key));

			ie.setIEDublinCore(dc);

			//	Add DNX to IE:
			Logger.debug("Adding DNX to IE in '" + aipPath + "' node: '" + node.getLabel() + "'");

			ie.setIeDnx(new DnxDocumentHelper(DnxDocumentFactory.getInstance().createDnxDocument()).getDocument());

			try
			{
				//	Special handling of retentionPolicy:
				String retentionPolicy = DB.getRosettaCodeIfExists(workspace, "retentionPolicy", (node.getDynamicMetadataValueForName("retentionPolicy")));
				ie.getAmdSecArray()[0].getRightsMDArray(0).getMdWrap().getXmlData().set(XmlObject.Factory.parse("<dnx xmlns='http://www.exlibrisgroup.com/dps/dnx'><section id='accessRightsPolicy'><record><key id='policyId'>" + retentionPolicy + "</key></record></section></dnx>"));

				Logger.debug("Adding retentionPolicy value: '" + retentionPolicy + "'");
			}
			catch (MetadataElementIsNotDefinedException ex)
			{
				//	Ignore this exception - in this case, the metadata element 'retentionPolicy' is not defined.
			}
			catch (MetadataElementIsNotAllowedException ex)
			{
				//	Ignore this exception - in this case, the metadata element 'retentionPolicy' is not allowed for this level.
			}

			try
			{
				//	Special handling of accessRestrictionPeriod (overwriting retentionPolicy if applicable):
				String accessRestrictionPeriod = DB.getRosettaCodeIfExists(workspace, "accessRestrictionPeriod", (node.getDynamicMetadataValueForName("accessRestrictionPeriod")));
				if (accessRestrictionPeriod != null)
				{
					ie.getAmdSecArray()[0].getRightsMDArray(0).getMdWrap().getXmlData().set(XmlObject.Factory.parse("<dnx xmlns='http://www.exlibrisgroup.com/dps/dnx'><section id='accessRightsPolicy'><record><key id='policyId'>" + accessRestrictionPeriod + "</key></record></section></dnx>"));

					Logger.debug("Adding accessRestrictionPeriod value: '" + accessRestrictionPeriod + "'");
				}
			}
			catch (MetadataElementIsNotDefinedException ex)
			{
				//	Ignore this exception - in this case, the metadata element 'accessRestrictionPeriod' is not defined.
			}
			catch (MetadataElementIsNotAllowedException ex)
			{
				//	Ignore this exception - in this case, the metadata element 'accessRestrictionPeriod' is not allowed for this level.
			}

			try
			{
				//	Special handling of retentionPeriod:
				String retentionPeriod = DB.getRosettaCodeIfExists(workspace, "retentionPeriod", (node.getDynamicMetadataValueForName("retentionPeriod")));
				ie.getAmdSecArray()[0].getTechMDArray(0).getMdWrap().getXmlData().set(XmlObject.Factory.parse("<dnx xmlns='http://www.exlibrisgroup.com/dps/dnx'><section id='retentionPolicy'><record><key id='policyId'>" + retentionPeriod + "</key></record></section></dnx>"));

				Logger.debug("Adding retentionPeriod value: '" + retentionPeriod + "'");
			}
			catch (MetadataElementIsNotDefinedException ex)
			{
				//	Ignore this exception - in this case, the metadata element 'retentionPeriod' is not defined.
			}
			catch (MetadataElementIsNotAllowedException ex)
			{
				//	Ignore this exception - in this case, the metadata element 'retentionPeriod' is not allowed for this level.
			}

			//	Special handling of characteristics/characteristicsNote:
			try
			{
				String characteristics = node.getDynamicMetadataValueForName("characteristics");
				String characteristicsNote = node.getDynamicMetadataValueForName("characteristicsNote");
				if (!(characteristics == null || characteristics.isEmpty()) && !(characteristicsNote == null || characteristicsNote.isEmpty())) {
					Logger.debug("Adding characteristics value: '" + characteristics + "', characteristicsNote value: '" + characteristicsNote + "'");
					ie.getAmdSecArray()[0].getTechMDArray(0).getMdWrap().getXmlData().set(XmlObject.Factory.parse("<dnx xmlns='http://www.exlibrisgroup.com/dps/dnx'><section id='generalIECharacteristics'><record><key id='status'>ACTIVE</key><key id='IEEntityType'>" + characteristics + "</key></record></section><section id='webHarvesting'><record><key id='primarySeedURL'>" + characteristicsNote + "</key></record></section></dnx>"));
				}
			}
			catch (MetadataElementIsNotDefinedException ex)
			{
				//	Ignore this exception - in this case, the metadata element 'retentionPeriod' is not defined.
			}
			catch (MetadataElementIsNotAllowedException ex)
			{
				//	Ignore this exception - in this case, the metadata element 'retentionPeriod' is not allowed for this level.
			}

			// Add node's metadata source (EAD-Type) to IE's amd section.
			// The EAD element contains all EAD elements from the root down to and including node:
			ie.setIeSourceMd(MdSecType.MdWrap.MDTYPE.EAD, XmlObject.Factory.parse(this.getEADElementPath(node).asXML()));

			if (node.isFile())
			{
				//	Single file:

				//	Add fileGrp:
				Logger.debug("Adding fileGrp to IE in '" + aipPath + "' node: '" + node.getLabel() + "'");
				FileGrp fGrp = ie.addNewFileGrp(Enum.UsageType.VIEW, Enum.PreservationType.PRESERVATION_MASTER);
				fGrp.setUSE(Enum.UsageType.VIEW.toString());

				// Add dnx - A new DNX is constructed and added on the file group
				Logger.debug("Adding DNX to fileGrp in '" + aipPath + "' node: '" + node.getLabel() + "'");
				DnxDocument fileGrpDnx = ie.getFileGrpDnx(fGrp.getID());
				this.fillFileGrpDnx(fileGrpDnx, node);
				ie.setFileDnx(fileGrpDnx, fGrp.getID());

				// add file and dnx metadata on file
				Logger.debug("Adding file to IE in '" + aipPath + "' node: '" + node.getLabel() + "'");
				FileType fileType = ie.addNewFile(fGrp, node.getMimeType(), node.getLabel(), null);	//	The 3rd argument comes into fileGrp.file.Flocat@xlink:href

				// add dnx - A new DNX is constructed and added on the file level
				Logger.debug("Adding DNX to file in '" + aipPath + "' node: '" + node.getLabel() + "'");
				DnxDocument fileDnx = ie.getFileDnx(fileType.getID());
				this.fillFileDnx(fileDnx, node, workspace);
				ie.setFileDnx(fileDnx, fileType.getID());

				// Copy the file we are ingesting into ".../content/streams":
				FileUtil.copyToFolderOverwriting(node.getAbsolutePathString(), streamsFilePath);
			}
			else
			{
				//	Folder (may contain multiple files):

				if (isFolderContainingMigratedFiles(node)) {
					Logger.debug("Is migrated IE, will create specific fileGrp for Representations");


					//	Add Original fileGrp:
					Logger.debug("Adding fileGrp to IE in '" + aipPath + "' node: '" + node.getLabel() + "'");
					FileGrp fGrpMaster = ie.addNewFileGrp(Enum.UsageType.VIEW, Enum.PreservationType.PRESERVATION_MASTER);
					fGrpMaster.setUSE(Enum.UsageType.VIEW.toString());

					// Add dnx - A new DNX is constructed and added on the file group
					Logger.debug("Adding DNX to fileGrp in '" + aipPath + "' node: '" + node.getLabel() + "'");
					DnxDocument fileGrpMasterDnx = ie.getFileGrpDnx(fGrpMaster.getID());
					this.fillFileGrpDnx(fileGrpMasterDnx, (NodeFile)node.getChildAt(0));
					ie.setFileDnx(fileGrpMasterDnx, fGrpMaster.getID());

					// add file and dnx metadata on file
					Logger.debug("Adding file to IE in '" + aipPath + "' node: '" + ((NodeFile)node.getChildAt(0)).getLabel() + "'");
					FileType fileTypeMaster = ie.addNewFile(fGrpMaster, ((NodeFile)node.getChildAt(0)).getMimeType(), ((NodeFile)node.getChildAt(0)).getLabel(), null);	//	The 3rd argument comes into fileGrp.file.Flocat@xlink:href

					// add dnx - A new DNX is constructed and added on the file level
					Logger.debug("Adding DNX to file in '" + aipPath + "' node: '" + ((NodeFile)node.getChildAt(0)).getLabel() + "'");
					DnxDocument fileDnxMaster = ie.getFileDnx(fileTypeMaster.getID());
					this.fillFileDnx(fileDnxMaster, (NodeFile)node.getChildAt(0), workspace);
					ie.setFileDnx(fileDnxMaster, fileTypeMaster.getID());

					// Copy the file we are ingesting into ".../content/streams":
					FileUtil.copyToFolderOverwriting(((NodeFile)node.getChildAt(0)).getAbsolutePathString(), streamsFilePath);


					//	Add Preservation fileGrp:
					Logger.debug("Adding fileGrp to IE in '" + aipPath + "' node: '" + node.getLabel() + "'");
					FileGrp fGrpDerived = ie.addNewFileGrp(Enum.UsageType.VIEW, Enum.PreservationType.MODIFIED_MASTER);
					fGrpDerived.setUSE(Enum.UsageType.VIEW.toString());

					// Add dnx - A new DNX is constructed and added on the file group
					Logger.debug("Adding DNX to fileGrp in '" + aipPath + "' node: '" + node.getLabel() + "'");
					DnxDocument fileGrpDerivedDnx = ie.getFileGrpDnx(fGrpDerived.getID());
					this.fillFileGrpDnx(fileGrpDerivedDnx, (NodeAbstract)node.getChildAt(1));
					ie.setFileDnx(fileGrpDerivedDnx, fGrpDerived.getID());

					// add file and dnx metadata on file
					Logger.debug("Adding file to IE in '" + aipPath + "' node: '" + ((NodeFile)node.getChildAt(1)).getLabel() + "'");
					FileType fileTypeDerived = ie.addNewFile(fGrpDerived, ((NodeFile)node.getChildAt(1)).getMimeType(), ((NodeFile)node.getChildAt(1)).getLabel(), null);	//	The 3rd argument comes into fileGrp.file.Flocat@xlink:href

					// add dnx - A new DNX is constructed and added on the file level
					Logger.debug("Adding DNX to file in '" + aipPath + "' node: '" + ((NodeFile)node.getChildAt(1)).getLabel() + "'");
					DnxDocument fileDnxDerived = ie.getFileDnx(fileTypeDerived.getID());
					this.fillFileDnx(fileDnxDerived, (NodeFile)node.getChildAt(1), workspace);
					ie.setFileDnx(fileDnxDerived, fileTypeDerived.getID());

					// Copy the file we are ingesting into ".../content/streams":
					FileUtil.copyToFolderOverwriting(((NodeFile)node.getChildAt(1)).getAbsolutePathString(), streamsFilePath);
				} else {
					Logger.debug("Creating a complex IE");
					
					//	check if complex IE should have multiple file groups because of migrations
					int migrationCount = 0;
					for (MutableTreeTableNode child: ((NodeFolder)node).getChildren()) {
						if (isFolderContainingMigratedFiles((NodeAbstract)child)) {
							migrationCount++;
						}
					}
					
					if (migrationCount > 0) {
						Logger.debug("Is migrated IE, will create specific fileGrp for Representations");


						//	Add Original fileGrp:
						Logger.debug("Adding fileGrp to IE in '" + aipPath + "' node: '" + node.getLabel() + "'");
						FileGrp fGrpMaster = ie.addNewFileGrp(Enum.UsageType.VIEW, Enum.PreservationType.PRESERVATION_MASTER);
						fGrpMaster.setUSE(Enum.UsageType.VIEW.toString());

						// Add dnx - A new DNX is constructed and added on the file group
						Logger.debug("Adding DNX to fileGrp in '" + aipPath + "' node: '" + node.getLabel() + "'");
						DnxDocument fileGrpMasterDnx = ie.getFileGrpDnx(fGrpMaster.getID());
						this.fillFileGrpDnx(fileGrpMasterDnx, node);
						ie.setFileDnx(fileGrpMasterDnx, fGrpMaster.getID());


						//	Insert all files and originals from children:
//						for (NodeAbstract subNode: ((NodeFolder)node).getDescendants())
						for (NodeFile subNode: ((NodeFolder)node).getDescendantFiles())
						{
							//	Only deal with nodes that are originals to migration actions
							if (subNode.getMigrationDerivedNode() == null)
								continue;

							//	Construct path between node and subNode (use "_" as path separator):
							StringBuilder filePathRelativeToIENodeStringBuilder = new StringBuilder(200);
							NodeAbstract parent = subNode;
							do
							{
								filePathRelativeToIENodeStringBuilder.insert(0, "_" + parent.getLabel());
								parent = (NodeAbstract)parent.getParent();
							}
							while (parent != node);
							//	Cut off leading "_":
							String filePathRelativeToIENode = filePathRelativeToIENodeStringBuilder.toString().substring(1);
	
							// Add subNode's metadata source (EAD-Type) to IE's amd section?
	//						ie.setIeSourceMd(MdSecType.MdWrap.MDTYPE.EAD, XmlObject.Factory.parse(subNode.getMyDMDSectionWithEAD().getElement().asXML()));
	
							// add file and dnx metadata on file
							Logger.debug("Adding file to complex IE in '" + aipPath + "' node: '" + filePathRelativeToIENode + "'");
							FileType fileType = ie.addNewFile(fGrpMaster, subNode.getMimeType(), filePathRelativeToIENode, null);	//	The 3rd argument comes into fileGrp.file.Flocat@xlin:href
	
							// add dnx - A new DNX is constructed and added on the file level
							DnxDocument fileDnx = ie.getFileDnx(fileType.getID());
							this.fillFileDnx(fileDnx, subNode, workspace);
							ie.setFileDnx(fileDnx, fileType.getID());
	
							//	Copy file:
							FileUtil.copyToOverwriting(subNode.getAbsolutePathString(), streamsFilePath + "/" + filePathRelativeToIENode);
						}


						//	Add Preservation fileGrp:
						Logger.debug("Adding fileGrp to IE in '" + aipPath + "' node: '" + node.getLabel() + "'");
						FileGrp fGrpDerived = ie.addNewFileGrp(Enum.UsageType.VIEW, Enum.PreservationType.MODIFIED_MASTER);
						fGrpDerived.setUSE(Enum.UsageType.VIEW.toString());

						// Add dnx - A new DNX is constructed and added on the file group
						Logger.debug("Adding DNX to fileGrp in '" + aipPath + "' node: '" + node.getLabel() + "'");
						DnxDocument fileGrpDerivedDnx = ie.getFileGrpDnx(fGrpDerived.getID());
						this.fillFileGrpDnx(fileGrpDerivedDnx, node);
						ie.setFileDnx(fileGrpDerivedDnx, fGrpDerived.getID());


						//	Insert all derivates from children:
						for (NodeFile subNode: ((NodeFolder)node).getDescendantFiles())
						{
							//	Only deal with nodes that are originals to migration actions
							if (subNode.getMigrationSourceNode() == null)
								continue;

							//	Construct path between node and subNode (use "_" as path separator):
							StringBuilder filePathRelativeToIENodeStringBuilder = new StringBuilder(200);
							NodeAbstract parent = subNode;
							do
							{
								filePathRelativeToIENodeStringBuilder.insert(0, "_" + parent.getLabel());
								parent = (NodeAbstract)parent.getParent();
							}
							while (parent != node);
							//	Cut off leading "_":
							String filePathRelativeToIENode = filePathRelativeToIENodeStringBuilder.toString().substring(1);
	
							// Add subNode's metadata source (EAD-Type) to IE's amd section?
	//						ie.setIeSourceMd(MdSecType.MdWrap.MDTYPE.EAD, XmlObject.Factory.parse(subNode.getMyDMDSectionWithEAD().getElement().asXML()));
	
							// add file and dnx metadata on file
							Logger.debug("Adding file to complex IE in '" + aipPath + "' node: '" + filePathRelativeToIENode + "'");
							FileType fileType = ie.addNewFile(fGrpDerived, subNode.getMimeType(), filePathRelativeToIENode, null);	//	The 3rd argument comes into fileGrp.file.Flocat@xlin:href
	
							// add dnx - A new DNX is constructed and added on the file level
							DnxDocument fileDnx = ie.getFileDnx(fileType.getID());
							this.fillFileDnx(fileDnx, subNode, workspace);
							ie.setFileDnx(fileDnx, fileType.getID());
	
							//	Copy file:
							FileUtil.copyToOverwriting(subNode.getAbsolutePathString(), streamsFilePath + "/" + filePathRelativeToIENode);
						}
					} else {
						//	Add single fileGrp:
						Logger.debug("Adding fileGrp to IE in '" + aipPath + "' node: '" + node.getLabel() + "'");
						FileGrp fGrp = ie.addNewFileGrp(Enum.UsageType.VIEW, Enum.PreservationType.PRESERVATION_MASTER);
						fGrp.setUSE(Enum.UsageType.VIEW.toString());
	
						// Add dnx - A new DNX is constructed and added on the file group
						Logger.debug("Adding DNX to fileGrp in '" + aipPath + "' node: '" + node.getLabel() + "'");
						DnxDocument fileGrpDnx = ie.getFileGrpDnx(fGrp.getID());
						this.fillFileGrpDnx(fileGrpDnx, node);
						ie.setFileDnx(fileGrpDnx, fGrp.getID());
						
						//	Insert all containing files within this folder, ignore containing folders:
						for (NodeFile subNode: ((NodeFolder)node).getDescendantFiles())
						{
							//	Construct path between node and subNode (use "_" as path separator):
							StringBuilder filePathRelativeToIENodeStringBuilder = new StringBuilder(200);
							NodeAbstract parent = subNode;
							do
							{
								filePathRelativeToIENodeStringBuilder.insert(0, "_" + parent.getLabel());
								parent = (NodeAbstract)parent.getParent();
							}
							while (parent != node);
							//	Cut off leading "_":
							String filePathRelativeToIENode = filePathRelativeToIENodeStringBuilder.toString().substring(1);
	
							// Add subNode's metadata source (EAD-Type) to IE's amd section?
	//						ie.setIeSourceMd(MdSecType.MdWrap.MDTYPE.EAD, XmlObject.Factory.parse(subNode.getMyDMDSectionWithEAD().getElement().asXML()));
	
							// add file and dnx metadata on file
							Logger.debug("Adding file to complex IE in '" + aipPath + "' node: '" + filePathRelativeToIENode + "'");
	
							FileType fileType = ie.addNewFile(fGrp, subNode.getMimeType(), filePathRelativeToIENode, null);	//	The 3rd argument comes into fileGrp.file.Flocat@xlin:href
	
							// add dnx - A new DNX is constructed and added on the file level
							DnxDocument fileDnx = ie.getFileDnx(fileType.getID());
							this.fillFileDnx(fileDnx, subNode, workspace);
							ie.setFileDnx(fileDnx, fileType.getID());
	
							//	Copy file:
							FileUtil.copyToOverwriting(subNode.getAbsolutePathString(), streamsFilePath + "/" + filePathRelativeToIENode);
						}
					}
				}
			}

			//	Finalization:

			//	Generate checksum:
//			ie.generateChecksum(streamsFilePath, Enum.FixityType.MD5.toString());
//			ie.updateSize(streamsFilePath);

			//	Create structmap:
			Logger.debug("Generating structMap in '" + aipPath + "' node: '" + node.getLabel() + "'");

			ie.generateStructMap(null, null, "Table of Contents");

			// Save IE into the content directory:
			Logger.debug("Saving '" + aipPath + "' node: '" + node.getLabel() + "'");

			com.exlibris.core.sdk.utils.FileUtil.writeFile(new File(ieFilePath), ie.toXML());

			// print the ie:
//			Tracer.trace(ie.toXML());
		}
		catch (java.lang.Exception ex)
		{
			errorMessages.add("MessageSubmitNodeExceptionCreatingIE '" + aipPath + "': " + ex.getMessage());
		}

		DB.logIECreationEnd(node, errorMessages.isEmpty()? IngestStatus.FEEDER_FINISHED: IngestStatus.FEEDER_ERROR, workspace, sipName);

		return errorMessages;
	}


	private void addDynamicMetadataValuesToDCElement(NodeAbstract node, String workspace, String mdName, DublinCore dc, String elementName)
	{
		Logger.debug("Mapping metadata: '" + mdName + "'");

		try
		{
			for (String s: node.getAllDynamicMetadataValuesForName(mdName))
				if (s != null && !s.isEmpty())
				{
					//	If there is a mapping in the DB Mapping Table for this mdName and mdValue, save the mapping, otherwise the mdValue itself:
					String value = DB.getRosettaCodeIfExists(workspace, mdName, s);
					dc.addElement(elementName, value);

					Logger.debug("Mapping: '" + mdName + "' to: '" + elementName + "' value: '" + value + "'");
				}
		}
		catch (Exception ex)
		{
			//	Ignore exception (e.g. when the dyn. metadata value is not accessable for this level) and just continue
		}
	}


	@SuppressWarnings("deprecation")
	private void fillFileGrpDnx(DnxDocument dnx, NodeAbstract node)
	{
		// Add dnx - A new DNX is constructed and added on the file group
		DnxDocumentHelper fileGrpDocumentHelper = new DnxDocumentHelper(dnx);
		//	What's set here comes into the rep's amd section:
		fileGrpDocumentHelper.getGeneralRepCharacteristics().setPreservationType(Enum.PreservationType.PRESERVATION_MASTER.toString());
		fileGrpDocumentHelper.getGeneralRepCharacteristics().setUsageType(Enum.UsageType.VIEW.toString());
		fileGrpDocumentHelper.getGeneralRepCharacteristics().setLabel(node.getLabel());
//		fileGrpDocumentHelper.getGeneralRepCharacteristics().setContentType("content type???");
//		fileGrpDocumentHelper.getGeneralRepCharacteristics().setContextType("context type???");
//		fileGrpDocumentHelper.getGeneralRepCharacteristics().setDeliveryPriority("1");
//		fileGrpDocumentHelper.getGeneralRepCharacteristics().setDerivedFromId("derived from id???");
//		fileGrpDocumentHelper.getGeneralRepCharacteristics().setDigitalOriginal("digital original???");
//		fileGrpDocumentHelper.getGeneralRepCharacteristics().setHardwareUsed("hardware used???");
//		fileGrpDocumentHelper.getGeneralRepCharacteristics().setOrderingSequence("2");
//		fileGrpDocumentHelper.getGeneralRepCharacteristics().setPhysicalCarrierMedia("physical carrier media???");
//		fileGrpDocumentHelper.getGeneralRepCharacteristics().setRepresentationCode("representation code???");
//		fileGrpDocumentHelper.getGeneralRepCharacteristics().setRepresentationEntityType("representation entity type???");
//		fileGrpDocumentHelper.getGeneralRepCharacteristics().setRepresentationOriginalName("representation original name???");
//		fileGrpDocumentHelper.getGeneralRepCharacteristics().setRevisionNumber("3");
//		fileGrpDocumentHelper.getGeneralRepCharacteristics().setTaskID("task id???");
	}


	private void fillFileDnx(DnxDocument dnx, NodeAbstract node, String workspace)
	{
		//	What's set here comes into the file's amd section:

		DnxDocumentHelper fileDocumentHelper = new DnxDocumentHelper(dnx);

		fileDocumentHelper.getGeneralFileCharacteristics().setLabel(node.getUnitTitle());
		fileDocumentHelper.getGeneralFileCharacteristics().setFileOriginalPath(workspace + "/" + node.getPathString());
		fileDocumentHelper.getGeneralFileCharacteristics().setFileOriginalName(node.getLabel());
//		fileDocumentHelper.getGeneralFileCharacteristics().setCompositionLevel("composition level???");
//		fileDocumentHelper.getGeneralFileCharacteristics().setFileCreationDate("file creation date???");
//		fileDocumentHelper.getGeneralFileCharacteristics().setFileEntityType("file entity type???");
//		fileDocumentHelper.getGeneralFileCharacteristics().setFileExtension("file extension???");
//		fileDocumentHelper.getGeneralFileCharacteristics().setFileLocation("file location???");
//		fileDocumentHelper.getGeneralFileCharacteristics().setFileLocationType("file location type???");
//		fileDocumentHelper.getGeneralFileCharacteristics().setFileModificationDate("file modification date???");
//		fileDocumentHelper.getGeneralFileCharacteristics().setFileOriginalID("file original id???");
//		fileDocumentHelper.getGeneralFileCharacteristics().setFormatLibraryId("format library id???");
//		fileDocumentHelper.getGeneralFileCharacteristics().setNote("note???");
//		fileDocumentHelper.getGeneralFileCharacteristics().setRiskLibraryIdentifiers("risk library ids???");

		fileDocumentHelper.setFileFixitys(Arrays.asList(new FileFixity[] {fileDocumentHelper.new FileFixity(null, null, node.getChecksumType(), node.getChecksum())}));
	}


	/**
	 * Create the path of EAD-Elements from the root to me.
	 * The root element might have to be adapted, as in the original EAD structure the top = root element is an archdesc.
	 * Here, the top element is an empty archdesc and the root element a c element as its only component.
	 * @param node
	 * @return
	 */
	private Element getEADElementPath(NodeAbstract node)
	{
		TreePath treePath = node.getTreePath();
		Element rootElement = (Element)((NodeAbstract)treePath.getPathComponent(0)).getMyDMDSectionWithEAD().getEADElement().clone();
		Element lastCElement = (Element)rootElement.selectSingleNode("EAD:archdesc/EAD:dsc/EAD:c");

		for(int i = 1; i < treePath.getPathCount(); i++)
		{
			NodeAbstract child = (NodeAbstract)treePath.getPathComponent(i);
			Element childCElement = (Element)child.getMyDMDSectionWithEAD().getEADElement().selectSingleNode("EAD:archdesc/EAD:dsc/EAD:c").clone();
			lastCElement.add(childCElement);
			lastCElement = childCElement;
		}

		for (Object submitStatusXMLNode: rootElement.selectNodes(".//EAD:name[@role='submitStatus']")) {
			((Node)submitStatusXMLNode).setText(SubmitStatus.Submitted.toString());
		}

		return rootElement;
	}

	//	--------		Interface			-------------------------------------------------------
	//	--------		Actions				-------------------------------------------------------
	//	--------		Business Ops		-------------------------------------------------------
	
	/**
	 * Check if the given node is a folder that was created in the context of a file migration.
	 * 
	 * @param node
	 * @return
	 */
	private boolean isFolderContainingMigratedFiles(NodeAbstract node) {
		if (node.getChildCount() == 2
				&& ((NodeAbstract)node.getChildAt(0)).isFile()
				&& ((NodeAbstract)node.getChildAt(1)).isFile()) {
			//	Is the first child the source for the second child?
			ch.docuteam.darc.premis.Object sourceObject = ((NodeAbstract)node.getChildAt(0)).getMyDigiprov().getLastObject();
			ch.docuteam.darc.premis.Object derivedObject = ((NodeAbstract)node.getChildAt(1)).getMyDigiprov().getLastObject();
			if (sourceObject.getRelatedObjectIdValue().equals(derivedObject.getId())
					&& sourceObject.getRelationshipType().equals("source")
					&& derivedObject.getRelationshipType().equals("derivation")) {
				return true;
			}
		}
		return false;
	}
	
	//	--------		Persistence			-------------------------------------------------------
	//	--------		Support				-------------------------------------------------------
	//	--------		Utilities			-------------------------------------------------------
	//	---------		Misc				-------------------------------------------------------
	//	--------		Debugging			-------------------------------------------------------
	//	---------		Temporary			-------------------------------------------------------

	//	TODO	This is temporary during development:
	private void _prepareTestRun(String submitPath) throws FileUtilExceptionListException
	{
		FileUtil.delete(submitPath);
	}

	//	===========================================================================================
	//	========	Inner Classes			=======================================================
	//	===========================================================================================

	//	--------		Submitting			-------------------------------------------------------
	
	static private abstract class DB
	{
		static private final String			_TrackingTableName = ".INGEST_TRACKING_RECORD_FEEDER ";
		static private final String			_CodeTableName = ".CODE_MAPPING ";

		static private String				TrackingTableName;
		static private String				CodeTableName;

		static private Connection			Connection;

		static private String				ExecutionId;


		static protected void initialize(String host, String port, String serviceName, String dbName, String userName, String userPassword, String executionId) throws SQLException
		{
			if (IsWithoutDB)		return;

			String connectionString = host + ":" + port + "/" + serviceName;

			Logger.debug("(" + executionId + ") DB Initialization: " + connectionString);

			TrackingTableName = dbName + _TrackingTableName;
			CodeTableName = dbName + _CodeTableName;

			Connection = DriverManager.getConnection(connectionString, userName, userPassword);

			ExecutionId = executionId;
		}


		static protected void close() throws SQLException
		{
			if (IsWithoutDB)		return;

			Logger.debug("(" + ExecutionId + ") DB Close");

			if (Connection != null)		Connection.close();
		}


		static protected boolean wasAlreadyIngested(String workspace, String sipName, NodeAbstract node) throws SQLException
		{
			if (IsWithoutDB)		return false;
			
			PreparedStatement pstmt = null;
			ResultSet resultSet = null;

			try
			{
				pstmt = Connection.prepareStatement(
					"select count(*) " +
					"from " + TrackingTableName +
					"where workspace_id = ? " +
					"	and sip_name = ? " +
					"	and amd_id = ?");
				pstmt.setString(1, workspace);
				pstmt.setString(2, sipName);
				pstmt.setString(3, node.getAdmId());
				resultSet = pstmt.executeQuery();

				return (resultSet.next() && resultSet.getInt(1) != 0);
			}
			finally
			{
				if (resultSet != null)		resultSet.close();
				if (pstmt != null)			pstmt.close();
			}
		}


		static protected void logIECreationStart(NodeAbstract node, String submitRootPath, String workspace, String sipName, String submitTimestamp) throws SQLException
		{
			if (IsWithoutDB)		return;

			String relativeSubmitRootPath = workspace + File.separator + sipName + File.separator + node.getPathString();
			String doi = null;
			if (node.getPID() != null)
				doi = node.getPID().get(0);

			PreparedStatement pstmt = null;
			try {
				pstmt = Connection.prepareStatement(
					"insert into " + TrackingTableName + "(" +
						"amd_id, " +
						"workflow_execution_id, " +
						"workspace_id," +
						"workspace_path," +
						"submit_timestamp," +
						"source_path," +
						"create_dt," +
						"update_dt," +
						"sip_name," +
						"sip_path," +
						"sip_type," +
						"sip_status," +
						"doi" +
					") values (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)");
				pstmt.setString(1, node.getAdmId());
				pstmt.setString(2, ExecutionId);
				pstmt.setString(3, workspace);
				pstmt.setString(4, relativeSubmitRootPath);
				pstmt.setTimestamp(5, Timestamp.valueOf(submitTimestamp));
				pstmt.setString(6, relativeSubmitRootPath);
				pstmt.setTimestamp(7, new Timestamp(new Date().getTime()));
				pstmt.setTimestamp(8, new Timestamp(new Date().getTime()));
				pstmt.setString(9, sipName);
				pstmt.setString(10, submitRootPath);
				pstmt.setString(11, "IE");
				pstmt.setString(12, IngestStatus.FEEDER_STARTED.toString());
				pstmt.setString(13, doi);

				Logger.debug("(" + ExecutionId + ") Generating IE:  '" + workspace + "/" + sipName + "/" + node.getAdmId() + "/" + node.getLabel() + "': Start");
				
				pstmt.executeUpdate();
			}
			finally
			{
				if (pstmt != null)			pstmt.close();
			}
		}


		static protected void logIECreationEnd(NodeAbstract node, IngestStatus status, String workspace, String sipName) throws SQLException
		{
			if (IsWithoutDB)		return;

			PreparedStatement pstmt = null;
			try {
				pstmt = Connection.prepareStatement(
					"update " + TrackingTableName +
					"set " +
						"sip_status = ?, " +
						"update_dt = ? " +
					"where workspace_id = ? " +
					"and sip_name = ? " +
					"and amd_id = ?");
				pstmt.setString(1, status.toString());
				pstmt.setTimestamp(2, new Timestamp(new Date().getTime()));
				pstmt.setString(3, workspace);
				pstmt.setString(4, sipName);
				pstmt.setString(5, node.getAdmId());

				Logger.debug("(" + ExecutionId + ") Generating IE:  '" + workspace + "/" + sipName + "/" + node.getAdmId() + "/" + node.getLabel() + "': " + status);

				pstmt.executeUpdate();
			}
			finally
			{
				if (pstmt != null)			pstmt.close();
			}
		}


		static protected void logCollectionCreationStart(NodeAbstract node, String submitRootPath, String workspace, String sipName, String submitTimestamp) throws SQLException
		{
			if (IsWithoutDB)		return;

			//	Collections can be re-submitted (for example if additional child nodes have been created or selected for submission; shouldn't craete a DB entry again. 
			if (DB.wasAlreadyIngested(workspace, sipName, node)) {
				Logger.debug("(" + ExecutionId + ") DB entry for :  '" + workspace + "/" + sipName + "/" + node.getAdmId() + "/" + node.getLabel() + "' already exists and will not be re-created");
			} else {
				String relativeSubmitRootPath = workspace + File.separator + sipName + File.separator + node.getPathString();
				String doi = null;
				if (node.getPID() != null)
					doi = node.getPID().get(0);
				
				PreparedStatement pstmt = null;
				try {
					pstmt = Connection.prepareStatement(
						"insert into " + TrackingTableName + "(" +
							"amd_id, " +
							"workflow_execution_id, " +
							"workspace_id," +
							"workspace_path," +
							"submit_timestamp," +
							"source_path," +
							"create_dt," +
							"update_dt," +
							"sip_name," +
							"sip_path," +
							"sip_type," +
							"sip_status," +
							"doi" +
						") values (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)");
					pstmt.setString(1, node.getAdmId());
					pstmt.setString(2, ExecutionId);
					pstmt.setString(3, workspace);
					pstmt.setString(4, relativeSubmitRootPath);
					pstmt.setTimestamp(5, Timestamp.valueOf(submitTimestamp));
					pstmt.setString(6, relativeSubmitRootPath);
					pstmt.setTimestamp(7, new Timestamp(new Date().getTime()));
					pstmt.setTimestamp(8, new Timestamp(new Date().getTime()));
					pstmt.setString(9, sipName);
					pstmt.setString(10, submitRootPath);
					pstmt.setString(11, "Collection");
					pstmt.setString(12, IngestStatus.FEEDER_STARTED.toString());
					pstmt.setString(13, doi);
	
					Logger.debug("(" + ExecutionId + ") Generating Call:  '" + workspace + "/" + sipName + "/" + node.getAdmId() + "/" + node.getLabel() + "': Start");
	
					pstmt.executeUpdate();
				}
				finally
				{
					if (pstmt != null)			pstmt.close();
				}
			}
		}


		static protected void logCollectionCreationEnd(NodeAbstract node, IngestStatus status, String workspace, String sipName) throws SQLException
		{
			if (IsWithoutDB)		return;

			PreparedStatement pstmt = null;
			try {
				pstmt = Connection.prepareStatement(
					"update " + TrackingTableName +
					"set " +
						"sip_status = ?, " +
						"update_dt = ? " +
					"where workspace_id = ? " +
						"and sip_name = ? " +
						"and amd_id = ?");
				pstmt.setString(1, status.toString());
				pstmt.setTimestamp(2, new Timestamp(new Date().getTime()));
				pstmt.setString(3, workspace);
				pstmt.setString(4, sipName);
				pstmt.setString(5, node.getAdmId());

				Logger.debug("(" + ExecutionId + ") Generating Call:  '" + workspace + "/" + sipName + "/" + node.getAdmId() + "/" + node.getLabel() + "': " + status);

				pstmt.executeUpdate();
			}
			finally
			{
				if (pstmt != null)		pstmt.close();
			}
		}


		static protected void logAIPCreationFinalizeOK(String workspace, String sipName) throws SQLException
		{
			if (IsWithoutDB)		return;

			PreparedStatement pstmt = null;
			try {
				pstmt = Connection.prepareStatement(
					"update " + TrackingTableName +
					"set " +
						"sip_status = ?, " +
						"update_dt = ? " +
					"where workspace_id = ? " +
						"and sip_name = ?" +
						"and sip_status in (?, ?)");
				pstmt.setString(1, IngestStatus.FEEDER_SUBMITTED.toString());
				pstmt.setTimestamp(2, new Timestamp(new Date().getTime()));
				pstmt.setString(3, workspace);
				pstmt.setString(4, sipName);
				pstmt.setString(5, IngestStatus.FEEDER_FINISHED.toString());
				pstmt.setString(6, IngestStatus.FEEDER_ERROR.toString());

				Logger.debug("(" + ExecutionId + ") Finalizing SIP: '" + workspace + "/" + sipName + "'");

				pstmt.executeUpdate();
			}
			finally
			{
				if (pstmt != null)		pstmt.close();
			}
		}


		/**
		 * Return a map with all status entries.
		 * @param document
		 * @return A Map with admIds as the key and the corresponding status as the value.
		 * @throws SQLException
		 */
		static protected Map<String, IngestStatus> getAllStatus(String workspace, String sipName) throws SQLException
		{
			if (IsWithoutDB)		return null;

			PreparedStatement pstmt = null;
			ResultSet resultSet = null;
			Map<String, IngestStatus> mapIngestStatus = new HashMap<String, IngestStatus>();

			try
			{
				pstmt = Connection.prepareStatement(
						"select amd_id, sip_status " +
						"from " + TrackingTableName +
						"where workspace_id = ? " +
						"	and sip_name = ?");
				pstmt.setString(1, workspace);
				pstmt.setString(2, sipName);
				resultSet = pstmt.executeQuery();

				while (resultSet.next())	mapIngestStatus.put(resultSet.getString(1), IngestStatus.named(resultSet.getString(2)));
			}
			finally
			{
				if (resultSet != null)		resultSet.close();
				if (pstmt != null)			pstmt.close();
			}

			return mapIngestStatus;
		}

		/**
		 * Return a map with all status entries.
		 * @param document
		 * @return A Map with admIds as the key and the corresponding status as the value.
		 * @throws SQLException
		 */
		static protected Map<String, String> getAllEntityIDs(String workspace, String sipName) throws SQLException
		{
			if (IsWithoutDB)		return null;

			PreparedStatement pstmt = null;
			ResultSet resultSet = null;
			Map<String, String> mapEntityIDs= new HashMap<String, String>();

			try {
				pstmt = Connection.prepareStatement(
						"select amd_id, entity_id " +
						"from " + TrackingTableName +
						"where workspace_id = ? " +
						"	and sip_name = ?");
				pstmt.setString(1, workspace);
				pstmt.setString(2, sipName);
				resultSet = pstmt.executeQuery();

				while (resultSet.next())		mapEntityIDs.put(resultSet.getString(1), resultSet.getString(2));
			}
			finally
			{
				if (resultSet != null)		resultSet.close();
				if (pstmt != null)			pstmt.close();
			}

			return mapEntityIDs;
		}


		/**
		 * If there is a mapping of the workspace, code type, and code value to a rosetta code, return the rosetta code, otherwise the code value itself.
		 * @param workspace
		 * @param codeType
		 * @param codeFeeder
		 * @return
		 * @throws SQLException
		 */
		static protected String getRosettaCodeIfExists(String workspace, String codeType, String codeFeeder) throws SQLException
		{
			if (IsWithoutDB)		return codeFeeder;

			PreparedStatement pstmt = null;
			ResultSet resultSet = null;
			String codeRosetta = null;
			try
			{
				pstmt = Connection.prepareStatement(
						"select code_rosetta " +
						"from " + CodeTableName +
						"where ws_refcodeadmin = ? " +
						"	and code_type = ? " +
						"	and code_feeder = ?");
				pstmt.setString(1, workspace);
				pstmt.setString(2, codeType);
				pstmt.setString(3, codeFeeder);
				resultSet = pstmt.executeQuery();

				if (resultSet.next())
				{
					codeRosetta = (resultSet.getString(1));
					Logger.debug("Translating: '" + codeType + "' from: '" + codeFeeder + "' to: '" + codeRosetta + "'");
				}
			}
			finally
			{
				if (resultSet != null)		resultSet.close();
				if (pstmt != null)			pstmt.close();
			}

			return codeRosetta != null? codeRosetta: codeFeeder;
		}

		//	---------		Temporary			-------------------------------------------------------

		//	TODO	This is temporary during development:
		static private void _prepareTestRun(String workspace, String workspacePath, String sipName) throws SQLException
		{
			if (IsWithoutDB)		return;

			PreparedStatement pstmt = null;
			try {
				pstmt = Connection.prepareStatement(
					"delete from " + TrackingTableName +
					"where workspace_id = ? " +
					"	and workspace_path = ? " +
					"	and sip_name = ?");
				pstmt.setString(1, workspace);
				pstmt.setString(2, workspace + "/" + sipName + "/" + workspacePath);
				pstmt.setString(3, sipName);

				Logger.debug("(" + ExecutionId + ") Preparing for testrun AIP:  '" + workspace + "/" + sipName + "/" + workspacePath + "'");

				pstmt.executeUpdate();
			}
			finally
			{
				if (pstmt != null)		pstmt.close();
			}
		}


		//	ToDo: This is temporary during development:
		static private void _setAllStatus(IngestStatus status, String workspace, String sipName) throws SQLException, IllegalArgumentException
		{
			if (IsWithoutDB)		return;

			PreparedStatement pstmt = null;
			try {
				pstmt = Connection.prepareStatement(
					"update " + TrackingTableName +
					"set sip_status = ? " +
					"where workspace_id = ? " +
					"	and sip_name = ?");
				pstmt.setString(1, status.toString());
				pstmt.setString(2, workspace);
				pstmt.setString(3, sipName);

				Logger.debug("(" + ExecutionId + ") Setting status AIP:  '" + workspace + "/" + sipName + "': Status: '" + status + "'");

				pstmt.executeUpdate();
			}
			finally
			{
				if (pstmt != null)		pstmt.close();
			}
		}
	}



	private enum AIPType
	{
		IE, Collection, None
	}



	public enum IngestStatus
	{
		IN_PROGRESS, IN_HUMAN_STAGE, IN_TA, ACTIVE, DECLINED, REJECTED, FINISHED,
		FEEDER_STARTED, FEEDER_FINISHED, FEEDER_ERROR, FEEDER_SUBMITTED,
		UNKNOWN;


		/**
		 * Return the IngestStatus with this name. If none exists, return IngestStatus.UNKNOWN.
		 * @param status
		 * @return
		 */
		static public IngestStatus named(String status)
		{
			try
			{
				return valueOf(status);
			}
			catch (Exception ex)
			{
				return UNKNOWN;
			}
		}
	}

}
