package ch.docuteam.packer.gui.filePreview;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Cursor;
import java.awt.Image;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Vector;

import javax.imageio.ImageIO;
import javax.media.jai.JAI;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.JToolBar;
import javax.swing.SwingConstants;

import org.icepdf.ri.common.SwingController;
import org.icepdf.ri.common.SwingViewBuilder;

import com.sun.jimi.core.Jimi;

import ch.docuteam.converter.OOConverter;
import ch.docuteam.darc.mets.structmap.NodeAbstract;
import ch.docuteam.tools.file.FileUtil;
import ch.docuteam.tools.id.UniqueID;
import ch.docuteam.tools.os.SystemProcess;
import ch.docuteam.tools.out.Logger;
import ch.docuteam.tools.string.StringUtil;
import ch.docuteam.tools.translations.I18N;

public class FilePreviewPanel extends JPanel {

	private static final String				TempPDFFileNameTemplate = FileUtil.getTempFolder() + "/FileConverter/Temp_%s.pdf";
	private static final String				TempPNGFileNameTemplate = FileUtil.getTempFolder() + "/FileConverter/Temp_%s.png";

	private static final int				DefaultCacheSizeLimit = 100;

	private static final String				Tab = "&nbsp;&nbsp;&nbsp;&nbsp;";

	//	The temporary files:
	private static List<String>				TempPDFFileNames = new Vector<String>();
	private static List<String>				TempPNGFileNames = new Vector<String>();

	//	This is the cache:
	private static Map<NodeAbstract, FilePreviewPanel>
											Cache = new HashMap<NodeAbstract, FilePreviewPanel>();

	//	This is the default cacheSize limit - the actual limit can be changed using setCacheSizeLimit():
	private static int						CacheSizeLimit = DefaultCacheSizeLimit;

	//	This list contains the order in which the cache elements have been retrieved:
	private static List<NodeAbstract>		CacheHitOrder = new ArrayList<NodeAbstract>(CacheSizeLimit);

	private FilePreviewPanel(NodeAbstract node)
	{
		super(new BorderLayout());

		try
		{
			this.setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));

			if (node == null)
			{
				Logger.getLogger().debug("No item selected");

				this.add(new JPanel());
				return;
			}

			if (!node.fileExists() || !node.canRead())
			{
				Logger.getLogger().debug("Item '" + node.getLabel() + "' could not be read");

				JLabel previewNotAvailable = new JLabel(I18N.translate("LabelPreviewCantReadFile"), SwingConstants.CENTER);
				previewNotAvailable.setOpaque(true);
				previewNotAvailable.setBackground(Color.LIGHT_GRAY);
				this.add(previewNotAvailable);

				return;
			}

			String fileName = node.getFile().getAbsolutePath();
			FileType fileType = FileType.check(node);

			switch (fileType)
			{
				case Folder:
				{
					Logger.getLogger().debug("Item '" + node.getLabel() + "' identified as folder");

					StringBuilder pathString = new StringBuilder("<html>" + I18N.translate("LabelPreviewFolder") + "<br>");
					int indent = 0;
					for (String pathElement: StringUtil.split(node.getPathString(), "/"))
					{
						pathString.append("<br>");
						for (int i = 0; i < indent; i++)		pathString.append(Tab);
						pathString.append(pathElement);
						indent++;
					}
					pathString.append("</html>");

					JLabel previewFolder = new JLabel(pathString.toString());
					previewFolder.setHorizontalAlignment(SwingConstants.CENTER);
					previewFolder.setOpaque(true);
					this.add(new JScrollPane(previewFolder));
					break;
				}
				case Text:
				{
					Logger.getLogger().debug("Item '" + node.getLabel() + "' identified as text file");

					this.add(new JScrollPane(createTextAreaWithFile(fileName)));

					break;
				}
				case HTML:
				{
					Logger.getLogger().debug("Item '" + node.getLabel() + "' identified as HTML file");

					JLabel previewHTML = new JLabel();
					previewHTML.setVerticalAlignment(SwingConstants.TOP);
					previewHTML.setText(FileUtil.getFileContentAsString(fileName));
					this.add(new JScrollPane(previewHTML));
					break;
				}
				case PDF:
				{
					Logger.getLogger().debug("Item '" + node.getLabel() + "' identified as PDF file");

					this.add(new PDFPreviewer(this, fileName));
					break;
				}
				case OOConvertable:
				{
					Logger.getLogger().debug("Item '" + node.getLabel() + "' identified as OO file");

					if (OOConverter.shallNotBeUsed())
					{
						JLabel previewNotAvailable = new JLabel(I18N.translate("LabelOOConverterNotAvailable"), SwingConstants.CENTER);
						previewNotAvailable.setOpaque(true);
						previewNotAvailable.setBackground(Color.LIGHT_GRAY);
						this.add(previewNotAvailable);
						break;
					}

					try
					{
						String tempPDFFileName = String.format(TempPDFFileNameTemplate, UniqueID.getString());
						OOConverter.convert2PDF(fileName, tempPDFFileName);

						TempPDFFileNames.add(tempPDFFileName);
						this.add(new PDFPreviewer(this, tempPDFFileName));
					}
					catch (Exception e)
					{
						e.printStackTrace();
						this.add(new JScrollPane(createPreviewError(e.toString())));
					}
					break;
				}
				case GraphicsNative:
				{
					Logger.getLogger().debug("Item '" + node.getLabel() + "' identified as native graphics file");

					this.add(new GraphicsPreviewer(this, new ImageIcon(fileName)));
					break;
				}
				case GraphicsImageIO:
				{
					Logger.getLogger().debug("Item '" + node.getLabel() + "' identified as ImageIO graphics file");

					try {
						this.add(new GraphicsPreviewer(this, new ImageIcon(ImageIO.read(new File(fileName)))));
					} catch (IOException e) {
						e.printStackTrace();
						this.add(new JScrollPane(createPreviewError(e.toString())));
					}
					break;
				}
				case GraphicsJAIConvertable:
				{
					Logger.getLogger().debug("Item '" + node.getLabel() + "' identified as JAI graphics file");

					try
					{
						this.add(new GraphicsPreviewer(this, new ImageIcon(JAI.create("fileload", fileName).getAsBufferedImage())));
					}
					catch (Exception e)
					{
						e.printStackTrace();
						this.add(new JScrollPane(createPreviewError(e.toString())));
					}
					break;
				}
				case GraphicsJAIConvertableImageRead:
				{
					Logger.getLogger().debug("Item '" + node.getLabel() + "' identified as JAI graphics file");

					try
					{
						this.add(new GraphicsPreviewer(this, new ImageIcon(JAI.create("ImageRead", fileName).getAsBufferedImage())));
					}
					catch (Exception e)
					{
						e.printStackTrace();
						this.add(new JScrollPane(createPreviewError(e.toString())));
					}
					break;
				}
				case GraphicsJIMIConvertable:
				{
					Logger.getLogger().debug("Item '" + node.getLabel() + "' identified as JIMI graphics file");

					try
					{
						this.add(new GraphicsPreviewer(this, new ImageIcon(Jimi.getImage(fileName))));
					}
					catch (Exception e)
					{
						e.printStackTrace();
						this.add(new JScrollPane(createPreviewError(e.toString())));
					}
					break;
				}
				case GraphicsImageMagickConvertable:
				{
					Logger.getLogger().debug("Item '" + node.getLabel() + "' identified as ImageMagick graphics file");

					try
					{
						String tempPNGFileName = String.format(TempPNGFileNameTemplate, UniqueID.getString());

						// TODO: Do not use absolute path to executable, use migration-config.xml
						String[] commandLine = new String[] {"/opt/local/bin/convert", "-compress", "none", fileName, new File(tempPNGFileName).getAbsolutePath()};
						int errorCode = SystemProcess.execute(commandLine);
						if (errorCode != 0)			throw new Exception("File conversion of '" + fileName + "' to PNG format returned error code: " + errorCode);

						TempPNGFileNames.add(tempPNGFileName);
						this.add(new GraphicsPreviewer(this, new ImageIcon(tempPNGFileName)));
					}
					catch (Exception e)
					{
						e.printStackTrace();
						this.add(new JScrollPane(createPreviewError(e.toString())));
					}
					break;
				}
				case Unknown:
				default:
				{
					Logger.getLogger().debug("Item '" + node.getLabel() + "' could not identify type");

					this.add(new UnknownFormatPreviewer(fileName));
					break;
				}
			}
		}
		finally
		{
			this.validate();
			this.repaint();

			this.setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
		}
	}

	/**
	 * Create a new FilePreviewPanel for the node passed, or take it from the cache if it is already in there:
	 */
	public static FilePreviewPanel create(NodeAbstract node, JPanel container)
	{
		//	If cache is turned off:
		if (CacheSizeLimit == 0)		return new FilePreviewPanel(node);

		if (Cache.containsKey(node))
		{
			//	Cached:
			Logger.getLogger().info("Cache hit: " + node.getLabel());

			//	Remove this node from the CacheHitOrder list, wherever it is (it will be inserted at the beginning of the list later):
			CacheHitOrder.remove(CacheHitOrder.indexOf(node));
		}
		else
		{
			//	Not yet cached:
			if (CacheHitOrder.size() >= CacheSizeLimit)
			{
				//	Cache limit size reached: remove "oldest" node:
				Logger.getLogger().debug("Dropped from cache: " + CacheHitOrder.get(0).getLabel());

				Cache.remove(CacheHitOrder.get(0));
				CacheHitOrder.remove(0);
			}

			//	Add node to the cache:
			Logger.getLogger().debug("Insert into cache: " + node.getLabel());

			Cache.put(node, new FilePreviewPanel(node));
		}

		//	Put this node at the beginning of the CacheHitOrder list:
		CacheHitOrder.add(node);

		Logger.getLogger().debug("Cache size: " + Cache.size() + "/" + CacheSizeLimit);

		return Cache.get(node);
	}


	public static void setCacheSizeLimit(int newCacheSizeLimit)
	{
		if (newCacheSizeLimit < 0)		throw new IllegalArgumentException("Cache size limit can not be set to a negative value");

		Logger.getLogger().debug("Setting FilePreviewer cache size to " + newCacheSizeLimit);

		if (newCacheSizeLimit == 0)
		{
			Cache.clear();
			CacheHitOrder.clear();
		}
		else if (newCacheSizeLimit < CacheSizeLimit)
		{
			//	Remove "oldest" nodes:
			for (int i = 0; i < Math.min((CacheSizeLimit - newCacheSizeLimit), CacheHitOrder.size()); i++)
			{
				Cache.remove(CacheHitOrder.get(0));
				CacheHitOrder.remove(0);
			}
		}

		CacheSizeLimit = newCacheSizeLimit;
	}


	public static int getCacheSizeLimit()
	{
		return CacheSizeLimit;
	}


	public static void cleanupTemporaryFiles()
	{
		PDFPreviewer.cleanup();

		Logger.getLogger().debug("PNGs to cleanup: " + TempPNGFileNames.size());
		for (String s: TempPNGFileNames)
		{
			Logger.getLogger().debug("Cleaning up: " + s);

			try
			{
				FileUtil.setWritable(s);
			}
			catch (Throwable ex)
			{
				// TODO Auto-generated catch block
				ex.printStackTrace();
			}
			try
			{
				FileUtil.deleteOnExit(s);
			}
			catch (Throwable ex)
			{
				// TODO Auto-generated catch block
				ex.printStackTrace();
			}
		}

		Logger.getLogger().debug("PDFs to cleanup: " + TempPDFFileNames.size());
		for (String s: TempPDFFileNames)
		{
			Logger.getLogger().debug("Cleaning up: " + s);

			try
			{
				FileUtil.setWritable(s);
			}
			catch (Throwable ex)
			{
				// TODO Auto-generated catch block
				ex.printStackTrace();
			}
			try
			{
				FileUtil.deleteOnExit(s);
			}
			catch (Throwable ex)
			{
				// TODO Auto-generated catch block
				ex.printStackTrace();
			}
		}
	}

	private static JTextArea createPreviewError(String text)
	{
		//	I need this JTextArea only in case of an error:
		JTextArea previewError = new JTextArea();
		previewError.setEditable(false);
		previewError.setLineWrap(true);
		previewError.setWrapStyleWord(true);
		previewError.setBackground(Color.YELLOW);
		previewError.setTabSize(4);
		previewError.setText(text);
		previewError.setCaretPosition(0);

		return previewError;
	}


	private static JTextArea createTextAreaWithFile(String fileName)
	{
		JTextArea previewText = new JTextArea();
		previewText.setEditable(false);
		previewText.setLineWrap(true);
		previewText.setWrapStyleWord(true);
		previewText.setTabSize(4);
		previewText.setText(FileUtil.getFileContentAsString(fileName));
		previewText.setCaretPosition(0);

		return previewText;
	}


	private static float zoomFactorToScaleProportionally(int imageWidth, int imageHeight, int displayWidth, int displayHeight)
	{
		return Math.min((float)displayWidth/imageWidth, (float)displayHeight/imageHeight);
	}

	private static enum FileType
	{
		Folder,
		Text, HTML, PDF, OOConvertable,
		GraphicsNative, GraphicsImageIO, GraphicsJAIConvertable, GraphicsJAIConvertableImageRead, GraphicsJIMIConvertable, GraphicsImageMagickConvertable,
		Unknown;

		private static FileType check(NodeAbstract node)
		{
			if (node.isFolder())						return Folder;

			//	First PUID:
			String formatKey = node.getFormatKey();
			     if (formatKey == null)					;										//	Just continue
			else if (formatKey.equals("x-fmt/111"))		return Text;							//	Text
			else if (formatKey.equals("x-fmt/18"))		return Text;							//	CSV
			else if (formatKey.equals("x-fmt/22"))		return Text;							//	ASCII
			else if (formatKey.equals("x-fmt/283"))		return Text;							//	ASCII
			else if (formatKey.equals("fmt/96"))		return HTML;							//	HTML
			else if (formatKey.equals("fmt/97"))		return HTML;							//	HTML
			else if (formatKey.equals("fmt/98"))		return HTML;							//	HTML
			else if (formatKey.equals("fmt/99"))		return HTML;							//	HTML
			else if (formatKey.equals("fmt/100"))		return HTML;							//	HTML
			else if (formatKey.equals("fmt/102"))		return HTML;							//	XHTML
			else if (formatKey.equals("fmt/103"))		return HTML;							//	XHTML
			else if (formatKey.equals("fmt/471"))		return HTML;							//	HTML
			else if (formatKey.equals("fmt/14"))		return PDF;								//	PDF
			else if (formatKey.equals("fmt/15"))		return PDF;								//	PDF
			else if (formatKey.equals("fmt/16"))		return PDF;								//	PDF
			else if (formatKey.equals("fmt/17"))		return PDF;								//	PDF
			else if (formatKey.equals("fmt/18"))		return PDF;								//	PDF
			else if (formatKey.equals("fmt/19"))		return PDF;								//	PDF
			else if (formatKey.equals("fmt/20"))		return PDF;								//	PDF
			else if (formatKey.equals("fmt/95"))		return PDF;								//	PDF/A
			else if (formatKey.equals("fmt/354"))		return PDF;								//	PDF/A
			else if (formatKey.equals("fmt/144"))		return PDF;								//	PDF/X
			else if (formatKey.equals("fmt/145"))		return PDF;								//	PDF/X
			else if (formatKey.equals("fmt/146"))		return PDF;								//	PDF/X
			else if (formatKey.equals("fmt/147"))		return PDF;								//	PDF/X
			else if (formatKey.equals("fmt/148"))		return PDF;								//	PDF/X
			else if (formatKey.equals("fmt/157"))		return PDF;								//	PDF/X
			else if (formatKey.equals("fmt/158"))		return PDF;								//	PDF/X
			else if (formatKey.equals("fmt/41"))		return GraphicsNative;					//	JPG
			else if (formatKey.equals("fmt/42"))		return GraphicsNative;					//	JPG
			else if (formatKey.equals("fmt/43"))		return GraphicsNative;					//	JPG
			else if (formatKey.equals("fmt/44"))		return GraphicsNative;					//	JPG
			else if (formatKey.equals("x-fmt/390"))		return GraphicsNative;					//	JPG
			else if (formatKey.equals("x-fmt/391"))		return GraphicsNative;					//	JPG
			else if (formatKey.equals("x-fmt/398"))		return GraphicsNative;					//	JPG
			else if (formatKey.equals("fmt/3"))			return GraphicsNative;					//	GIF
			else if (formatKey.equals("fmt/4"))			return GraphicsNative;					//	GIF
			else if (formatKey.equals("fmt/11"))		return GraphicsNative;					//	PNG
			else if (formatKey.equals("fmt/12"))		return GraphicsNative;					//	PNG
			else if (formatKey.equals("fmt/13"))		return GraphicsNative;					//	PNG
			else if (formatKey.equals("fmt/114"))		return GraphicsJAIConvertable;			//	BMP
			else if (formatKey.equals("fmt/115"))		return GraphicsJAIConvertable;			//	BMP
			else if (formatKey.equals("fmt/116"))		return GraphicsJAIConvertable;			//	BMP
			else if (formatKey.equals("fmt/117"))		return GraphicsJAIConvertable;			//	BMP
			else if (formatKey.equals("fmt/118"))		return GraphicsJAIConvertable;			//	BMP
			else if (formatKey.equals("fmt/119"))		return GraphicsJAIConvertable;			//	BMP
			else if (formatKey.equals("x-fmt/270"))		return GraphicsJAIConvertable;			//	BMP
			else if (formatKey.equals("fmt/10"))		return GraphicsJAIConvertable;			//	TIFF
			else if (formatKey.equals("fmt/353"))		return GraphicsJAIConvertable;			//	TIFF
			else if (formatKey.equals("fmt/151"))		return GraphicsImageIO;	//	JPG2000
			else if (formatKey.equals("x-fmt/392"))		return GraphicsJAIConvertableImageRead;	//	JPG2000
			else if (formatKey.equals("x-fmt/92"))		return GraphicsJIMIConvertable;			//	PSD
			else if (formatKey.equals("fmt/122"))		return GraphicsImageMagickConvertable;	//	EPS
			else if (formatKey.equals("fmt/123"))		return GraphicsImageMagickConvertable;	//	EPS
			else if (formatKey.equals("fmt/124"))		return GraphicsImageMagickConvertable;	//	EPS
			else if (formatKey.equals("fmt/45"))		return OOConvertable;					//	RTF
			else if (formatKey.equals("fmt/50"))		return OOConvertable;					//	RTF
			else if (formatKey.equals("fmt/51"))		return OOConvertable;					//	RTF
			else if (formatKey.equals("fmt/52"))		return OOConvertable;					//	RTF
			else if (formatKey.equals("fmt/53"))		return OOConvertable;					//	RTF
			else if (formatKey.equals("fmt/355"))		return OOConvertable;					//	RTF
			else if (formatKey.equals("fmt/37"))		return OOConvertable;					//	MS Word
			else if (formatKey.equals("fmt/38"))		return OOConvertable;					//	MS Word
			else if (formatKey.equals("fmt/39"))		return OOConvertable;					//	MS Word
			else if (formatKey.equals("fmt/40"))		return OOConvertable;					//	MS Word
			else if (formatKey.equals("x-fmt/129"))		return OOConvertable;					//	MS Word
			else if (formatKey.equals("x-fmt/131"))		return OOConvertable;					//	MS Word
			else if (formatKey.equals("x-fmt/273"))		return OOConvertable;					//	MS Word
			else if (formatKey.equals("x-fmt/274"))		return OOConvertable;					//	MS Word
			else if (formatKey.equals("x-fmt/275"))		return OOConvertable;					//	MS Word
			else if (formatKey.equals("x-fmt/276"))		return OOConvertable;					//	MS Word
			else if (formatKey.equals("x-fmt/473"))		return OOConvertable;					//	MS Word
			else if (formatKey.equals("fmt/412"))		return OOConvertable;					//	MS Word
			else if (formatKey.equals("fmt/42"))		return OOConvertable;					//	Wordperfect
			else if (formatKey.equals("fmt/43"))		return OOConvertable;					//	Wordperfect
			else if (formatKey.equals("fmt/44"))		return OOConvertable;					//	Wordperfect
			else if (formatKey.equals("fmt/55"))		return OOConvertable;					//	MS Excel
			else if (formatKey.equals("fmt/56"))		return OOConvertable;					//	MS Excel
			else if (formatKey.equals("fmt/57"))		return OOConvertable;					//	MS Excel
			else if (formatKey.equals("fmt/59"))		return OOConvertable;					//	MS Excel
			else if (formatKey.equals("fmt/60"))		return OOConvertable;					//	MS Excel
			else if (formatKey.equals("fmt/61"))		return OOConvertable;					//	MS Excel
			else if (formatKey.equals("fmt/62"))		return OOConvertable;					//	MS Excel
			else if (formatKey.equals("fmt/172"))		return OOConvertable;					//	MS Excel
			else if (formatKey.equals("fmt/173"))		return OOConvertable;					//	MS Excel
			else if (formatKey.equals("fmt/174"))		return OOConvertable;					//	MS Excel
			else if (formatKey.equals("fmt/175"))		return OOConvertable;					//	MS Excel
			else if (formatKey.equals("fmt/176"))		return OOConvertable;					//	MS Excel
			else if (formatKey.equals("fmt/177"))		return OOConvertable;					//	MS Excel
			else if (formatKey.equals("fmt/178"))		return OOConvertable;					//	MS Excel
			else if (formatKey.equals("fmt/214"))		return OOConvertable;					//	MS Excel
			else if (formatKey.equals("fmt/445"))		return OOConvertable;					//	MS Excel
			else if (formatKey.equals("fmt/125"))		return OOConvertable;					//	MS Powerpoint
			else if (formatKey.equals("fmt/126"))		return OOConvertable;					//	MS Powerpoint
			else if (formatKey.equals("x-fmt/88"))		return OOConvertable;					//	MS Powerpoint
			else if (formatKey.equals("fmt/179"))		return OOConvertable;					//	MS Powerpoint
			else if (formatKey.equals("fmt/180"))		return OOConvertable;					//	MS Powerpoint
			else if (formatKey.equals("fmt/181"))		return OOConvertable;					//	MS Powerpoint
			else if (formatKey.equals("fmt/182"))		return OOConvertable;					//	MS Powerpoint
			else if (formatKey.equals("fmt/215"))		return OOConvertable;					//	MS Powerpoint
			else if (formatKey.equals("fmt/111"))		return OOConvertable;					//	MS OFfice
			else if (formatKey.equals("fmt/189"))		return OOConvertable;					//	MS Open Office

			//	Then check MIME Type:
			String mimeType = node.getMimeType();
			     if (mimeType == null)					;										//	Just continue
			else if (mimeType.equals("text/plain"))		return Text;							//	Text
			else if (mimeType.equals("text/xml"))		return Text;							//	XML
			else if (mimeType.equals("text/html"))		return HTML;							//	HTML
			else if (mimeType.equals("application/xhtml+xml"))
														return HTML;							//	XHTML
			else if (mimeType.equals("application/pdf"))return PDF;								//	PDF
			else if (mimeType.equals("image/jpeg"))		return GraphicsNative;					//	JPG
			else if (mimeType.equals("image/gif"))		return GraphicsNative;					//	GIF
			else if (mimeType.equals("image/png"))		return GraphicsNative;					//	PNG
			else if (mimeType.equals("application/postscript"))
														return GraphicsImageMagickConvertable;	//	EPS
			else if (mimeType.equals("application/rtf"))return OOConvertable;					//	RTF

			//	Finally check file extension:
			String extension = FileUtil.asFileNameExtension(node.getLabel());
			if (extension != null)
			{
				extension = extension.toLowerCase();
				     if (extension.equals("txt"))		return Text;							//	Text
				else if (extension.equals("xml"))		return Text;							//	XML
				else if (extension.equals("properties"))return Text;							//	Properties
				else if (extension.equals("props"))		return Text;							//	Properties
				else if (extension.equals("config"))	return Text;							//	Configuration
				else if (extension.equals("conf"))		return Text;							//	Configuration
				else if (extension.equals("log"))		return Text;							//	Log
				else if (extension.equals("html"))		return HTML;							//	HTML
				else if (extension.equals("htm"))		return HTML;							//	HTML
				else if (extension.equals("pdf"))		return PDF;								//	PDF
				else if (extension.equals("gif"))		return GraphicsNative;					//	GIF
				else if (extension.equals("png"))		return GraphicsNative;					//	PNG
				else if (extension.equals("jpeg"))		return GraphicsNative;					//	JPG
				else if (extension.equals("jpg"))		return GraphicsNative;					//	JPG
				else if (extension.equals("bmp"))		return GraphicsJAIConvertable;			//	BMP
				else if (extension.equals("tiff"))		return GraphicsJAIConvertable;			//	TIF
				else if (extension.equals("tif"))		return GraphicsJAIConvertable;			//	TIF
				else if (extension.equals("psd"))		return GraphicsJIMIConvertable;			//	PSD
				else if (extension.equals("eps"))		return GraphicsImageMagickConvertable;	//	EPS
				else if (extension.equals("rtf"))		return OOConvertable;					//	RTF
				else if (extension.equals("doc"))		return OOConvertable;					//	MS Word
				else if (extension.equals("docx"))		return OOConvertable;					//	MS X Word
				else if (extension.equals("xls"))		return OOConvertable;					//	MS Excel
				else if (extension.equals("xlsx"))		return OOConvertable;					//	MS X Excel
				else if (extension.equals("ppt"))		return OOConvertable;					//	MS PowerPoint
				else if (extension.equals("pptx"))		return OOConvertable;					//	MS X PowerPoint
			}

			return Unknown;
		}
	}


	/**
	 * The coding of this class follows closely the example
	 * ViewerComponentExample.java which is distributed with the ICEpdf
	 * library.
	 *
	 * @author andreas
	 *
	 */
	private static class PDFPreviewer extends JPanel
	{
		//	I have to close all open PDFPreviewers, otherwise the files being displayed remain locked!
		private static List<PDFPreviewer>		AllPDFPreviewers = new Vector<PDFPreviewer>(100);

		private SwingController					controller;

		private PDFPreviewer(FilePreviewPanel parent, String fileName)
		{
			super(new BorderLayout());
			AllPDFPreviewers.add(this);

			this.controller = new SwingController();
			this.controller.setIsEmbeddedComponent(true);

			DocuteamPackerSwingViewBuilder factory = new DocuteamPackerSwingViewBuilder(this.controller);

			JToolBar toolBar = factory.buildCompleteToolBar(true);

			if (toolBar != null)
				this.add(toolBar, BorderLayout.NORTH);

			this.add(this.controller.getDocumentViewController().getViewContainer(), BorderLayout.CENTER);

			JPanel statusPanel = factory.buildStatusPanel();
			if (statusPanel != null)
				this.add(statusPanel, BorderLayout.SOUTH);

			this.controller.openDocument(fileName);
			// ToDo This somehow doesn't have the intended effect yet:
			this.controller.setPageFitMode(org.icepdf.core.views.DocumentViewController.PAGE_FIT_ACTUAL_SIZE, true);
		}

		/**
		 * The reason for using a subclass is to be able to use the
		 * sometimes private methods of its superclass.
		 *
		 * @author andreas
		 *
		 */
		protected class DocuteamPackerSwingViewBuilder extends SwingViewBuilder {

			public DocuteamPackerSwingViewBuilder(SwingController c) {
				super(c);
			}

			/**
			 * This reduces the toolbar to the necessary components for the preview.
			 */
			@Override
			public JToolBar buildCompleteToolBar(boolean embeddableComponent) {
				JToolBar toolbar = new JToolBar();
				this.commonToolBarSetup(toolbar, true);

				// Build the main set of toolbars based on the property file configuration
				this.addToToolBar(toolbar, this.buildPageNavigationToolBar());
				this.addToToolBar(toolbar, this.buildZoomToolBar());
				this.addToToolBar(toolbar, this.buildFitToolBar());

				// Set the toolbar back to null if no components were added
				// The result of this will properly disable the necessary menu items for controlling the toolbar
				if (toolbar.getComponentCount() == 0) {
					toolbar = null;
				}

				if ((this.viewerController != null) && (toolbar != null))
					this.viewerController.setCompleteToolBar(toolbar);

				return toolbar;
			}
		}


		/**
		 * If I don't close and dispose all controllers, the PDF files being displayed remain locked!
		 */
		public static void cleanup()
		{
			for (PDFPreviewer p: AllPDFPreviewers)
			{
				p.controller.closeDocument();
				p.controller.dispose();
			}
		}
	}


	private static class GraphicsPreviewer extends JLabel
	{
		private FilePreviewPanel	parent;

		private ImageIcon			previewGraphicsImageIconOriginalSize;


		private GraphicsPreviewer(FilePreviewPanel parent, ImageIcon imageIcon)
		{
			super();

			this.parent = parent;

			this.setVerticalAlignment(SwingConstants.CENTER);
			this.setHorizontalAlignment(SwingConstants.CENTER);
			this.addComponentListener(new ComponentAdapter() {
					@Override public void componentResized(ComponentEvent e) { GraphicsPreviewer.this.wasResized(); }});

			this.previewGraphicsImageIconOriginalSize = imageIcon;
		}


		private void wasResized()
		{
			this.setIcon(this.scaleProportionallyTo(this.previewGraphicsImageIconOriginalSize, this.parent.getWidth(), this.parent.getHeight()));
			this.repaint();
		}


		private ImageIcon scaleProportionallyTo(ImageIcon originalImageIcon, int displayWidth, int displayHeight)
		{
			int imageWidth = originalImageIcon.getIconWidth();
			int imageHeight = originalImageIcon.getIconHeight();

			float zoomFactor = zoomFactorToScaleProportionally(imageWidth, imageHeight, displayWidth, displayHeight);
			if (zoomFactor >= 1)		return originalImageIcon;

			int zoomedImageWidth = (int)(zoomFactor * imageWidth);
			int zoomedImageHeight = (int)(zoomFactor * imageHeight);

			return new ImageIcon(originalImageIcon.getImage().getScaledInstance(zoomedImageWidth, zoomedImageHeight, Image.SCALE_FAST));
		}
	}



	private static class UnknownFormatPreviewer extends JPanel
	{
		private UnknownFormatPreviewer(final String fileName)
		{
			super(new BorderLayout());

			JButton previewNotAvailableButton = new JButton(I18N.translate("LabelPreviewNotAvailable"));
			previewNotAvailableButton.addActionListener(new ActionListener()
				{ @Override public void actionPerformed(ActionEvent e) { UnknownFormatPreviewer.this.showFileContentsAsText(fileName); }});
			previewNotAvailableButton.setBorder(null);
			previewNotAvailableButton.setOpaque(true);
			previewNotAvailableButton.setHorizontalAlignment(SwingConstants.CENTER);
			previewNotAvailableButton.setBackground(Color.LIGHT_GRAY);

			this.add(previewNotAvailableButton, BorderLayout.CENTER);
		}


		private void showFileContentsAsText(String fileName)
		{
			this.removeAll();
			this.add(new JScrollPane(createTextAreaWithFile(fileName)), BorderLayout.CENTER);
			this.validate();		//	Re-render this view
		}
	}


}
