/**
 *	Copyright (C) 2011-2016 Docuteam GmbH
 *
 *	This program is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 3
 *	as published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package ch.docuteam.packerETH.postactions;

import java.util.List;

import ch.docuteam.darc.exceptions.MetadataElementSetterPostActionException;
import ch.docuteam.darc.mdconfig.MetadataElement;
import ch.docuteam.darc.mdconfig.MetadataElementInstance;
import ch.docuteam.darc.mdconfig.MetadataElementSetterPostAction;
import ch.docuteam.darc.mets.structmap.NodeAbstract;
import ch.docuteam.darc.mets.structmap.NodeFolder;

/**
 * This setterPostAction adds the new value in all parents. If the parent has this value already set, skip this parent.
 *
 * IMPORTANT NOTE: If this dynamic metadata value is not repeatable for a parent, a MetadataElementCantAddException will be thrown.
 * The parents whose metadata value was changed until then, remain set.
 *
 * @author denis
 *
 */
public class MetadataElementSetterPostActionAddValueInParents implements MetadataElementSetterPostAction
{
	/* (non-Javadoc)
	 * @see ch.docuteam.darc.mdconfig.MetadataElementSetterPostAction#execute(java.lang.String, java.lang.String, ch.docuteam.darc.mets.structmap.NodeAbstract, ch.docuteam.darc.mdconfig.MetadataElement, int)
	 */
	@Override
	public void execute(String oldValue, String newValue, NodeAbstract context, MetadataElement metadataElement, int index)
			throws MetadataElementSetterPostActionException
	{
		String metadataAccessorName = metadataElement.getAccessorName();

		//	NOTE: This call is implicitly recursive, because sending the setter message "setDynamicMetadataValueForName()"
		//	to the parent triggers this check for the parent's parent (and so on).
		NodeFolder parent = (NodeFolder)context.getParent();
		if (parent != null)
		{
			try
			{
				if (newValue == null || newValue.isEmpty())
				{
					// Remove existing values only if newValue is null
					List<NodeAbstract> descendants = parent.getDescendants();
					for (NodeAbstract descendant : descendants)
					{
						List<String> existingValues = descendant.getAllDynamicMetadataValuesForName(metadataAccessorName);
						if (existingValues.contains(oldValue))
						{
							// Other descendants have this value, so do not remove value from parent
							return;
						}
					}
					// This value has not been found in descendants so remove existing value from parent
					int selectedInstanceIndex = -1;
					int selectedValueIndex = -1;
//					List<MetadataElementInstance> selectedInstances = new ArrayList<MetadataElementInstance>();
					List<MetadataElementInstance> instances = parent.getDynamicMetadataElementInstances();
					for (int instanceIndex = 0; instanceIndex < instances.size(); instanceIndex++)
					{
						MetadataElementInstance instance = instances.get(instanceIndex);
						if (instance.getLevelMetadataElement().getMetadataElement().getAccessorName().equals(metadataAccessorName))
						{
							
							selectedInstanceIndex = instanceIndex;
							List<String> values = instance.getLevelMetadataElement().getAllValuesFromNode(parent);
							for (int valueIndex = 0; valueIndex < values.size(); valueIndex++)
							{
								if (values.get(valueIndex).equals(oldValue))
								{
									selectedValueIndex = valueIndex;
									break;
								}
							}
						}
					}
					if (selectedValueIndex > -1)
					{
						MetadataElementInstance instance = instances.get(selectedInstanceIndex);
						instance.getLevelMetadataElement().setValueInNode(selectedValueIndex, newValue, parent);
//						if (instance.getLevelMetadataElement().getAllValuesFromNode(parent).size() > 1)
//						{
//							instance.getLevelMetadataElement().getAllValuesFromNode(parent).remove(selectedValueIndex);
//						}
					}
				}
				else
				{
					//	Does this parent already contain this value? If yes, skip it:
					List<String> existingValues = parent.getAllDynamicMetadataValuesForName(metadataAccessorName);
					if (!existingValues.contains(newValue))
					{
						//  Before adding look for empty value elements and use one of these if available
						if (existingValues.contains(""))
						{
							for (int valueIndex = 0; valueIndex < existingValues.size(); valueIndex++)
							{
								if (existingValues.get(valueIndex).equals(""))
								{
									parent.setDynamicMetadataValueForName(metadataAccessorName, valueIndex, newValue);
									break;
								}
							}
						}
						else
						{
							//	Add to the parent one instance of DynamicMetadataElementInstance and set its value:
							parent.addDynamicMetadataElementInstanceWithName(metadataAccessorName);
							//  existingValues has to be reloaded to reflect the actual count of values
							existingValues = parent.getAllDynamicMetadataValuesForName(metadataAccessorName);
							
							//	IMPORTANT NOTE: By now, the size of the list "existingValues" has increased by one!
							parent.setDynamicMetadataValueForName(metadataAccessorName, existingValues.size() - 1, newValue);
						}
					}
				}
			}
			catch (Exception ex)
			{
				throw new MetadataElementSetterPostActionException(ex);
			}
		}
	}

}
